from zipfile import ZipFile
from zipfile import ZIP_STORED
import json
import struct
from ast import literal_eval
import os
import mmap
import typing as tp

import numpy as np

from static_frame.core.util import PathSpecifier
from static_frame.core.util import DTYPE_OBJECT_KIND
from static_frame.core.util import list_to_tuple
from static_frame.core.util import EMPTY_TUPLE
from static_frame.core.util import NameType
from static_frame.core.container_util import ContainerMap

from static_frame.core.exception import ErrorNPYDecode
from static_frame.core.exception import ErrorNPYEncode

if tp.TYPE_CHECKING:
    import pandas as pd #pylint: disable=W0611 #pragma: no cover
    from static_frame.core.index_base import IndexBase #pylint: disable=W0611,C0412 #pragma: no cover
    from static_frame.core.frame import Frame #pylint: disable=W0611,C0412 #pragma: no cover




#-------------------------------------------------------------------------------

class NPYConverter:
    '''Optimized implementation based on numpy/lib/format.py
    '''
    MAGIC_PREFIX = b'\x93NUMPY' + bytes((1, 0)) # version 1.0
    MAGIC_LEN = len(MAGIC_PREFIX)
    ARRAY_ALIGN = 64
    STRUCT_FMT = '<H' # version 1.0
    STRUCT_FMT_SIZE = struct.calcsize(STRUCT_FMT)
    ENCODING = 'latin1' # version 1.0
    BUFFERSIZE_NUMERATOR = 16 * 1024 ** 2
    NDITER_FLAGS = ('external_loop', 'buffered', 'zerosize_ok')

    @classmethod
    def _header_encode(cls, header: str) -> bytes:
        '''
        Takes a string header, and attaches the prefix and padding to it.
        This is hard-coded to only use Version 3.0
        '''
        center = header.encode(cls.ENCODING)
        hlen = len(center) + 1

        padlen = cls.ARRAY_ALIGN - (
               (cls.MAGIC_LEN + cls.STRUCT_FMT_SIZE + hlen) % cls.ARRAY_ALIGN
               )
        prefix = cls.MAGIC_PREFIX + struct.pack(cls.STRUCT_FMT, hlen + padlen)
        postfix = b' ' * padlen + b'\n'
        return prefix + center + postfix

    @classmethod
    def to_npy(cls, file: tp.IO[bytes], array: np.ndarray) -> None:
        '''Write an NPY 3.0 file to the open, writeable, binary file given by ``file``.
        '''
        dtype = array.dtype
        if dtype.kind == DTYPE_OBJECT_KIND:
            raise ErrorNPYEncode('no support for object dtypes')
        if dtype.names is not None:
            raise ErrorNPYEncode('no support for structured arrays')
        if array.ndim == 0 or array.ndim > 2:
            raise ErrorNPYEncode('no support for ndim == 0 or greater than two.')

        flags = array.flags
        fortran_order = flags.f_contiguous

        # NOTE: do not know how to create array with itmesize 0, assume array.itemsize > 0
        # NOTE: derived numpy configuration
        buffersize = max(cls.BUFFERSIZE_NUMERATOR // array.itemsize, 1)

        header = f'{{"descr":"{dtype.str}","fortran_order":{fortran_order},"shape":{array.shape}}}'
        file.write(cls._header_encode(header))

        # this works but forces loading everything in memory
        # if flags.f_contiguous and not flags.c_contiguous:
        #     file.write(array.T.tobytes())
        # else:
        #     file.write(array.tobytes())

        # NOTE: this might be made more efficient by creating an ArrayKit function that extracts bytes directly, avoiding creating an array for each chunk.
        if fortran_order and not flags.c_contiguous:
            for chunk in np.nditer(
                    array,
                    flags=cls.NDITER_FLAGS,
                    buffersize=buffersize,
                    order='F',
                    ):
                file.write(chunk.tobytes('C'))
        else:
            for chunk in np.nditer(
                    array,
                    flags=cls.NDITER_FLAGS,
                    buffersize=buffersize,
                    order='C',
                    ):
                file.write(chunk.tobytes('C'))



    @classmethod
    def _header_decode(cls,
            file: tp.IO[bytes],
            ) -> tp.Tuple[np.dtype, bool, tp.Tuple[int, ...]]:
        '''Extract and decode the header.
        '''
        length_size = file.read(cls.STRUCT_FMT_SIZE)
        length_header = struct.unpack(cls.STRUCT_FMT, length_size)[0]
        header = file.read(length_header).decode(cls.ENCODING)
        dtype_str, fortran_order, shape = literal_eval(header).values()
        return np.dtype(dtype_str), fortran_order, shape

    @classmethod
    def from_npy(cls,
            file: tp.IO[bytes],
            memory_map: bool = False,
            ) -> tp.Tuple[np.ndarray, tp.Optional[mmap.mmap]]:
        '''Read an NPY 1.0 file.
        '''
        if cls.MAGIC_PREFIX != file.read(cls.MAGIC_LEN):
            raise ErrorNPYDecode('Invalid NPY header found.')

        dtype, fortran_order, shape = cls._header_decode(file)
        if dtype.kind == DTYPE_OBJECT_KIND:
            raise ErrorNPYDecode('no support for object dtypes')

        ndim = len(shape)
        if ndim == 1:
            size = shape[0]
        elif ndim == 2:
            size = shape[0] * shape[1]
        else:
            raise ErrorNPYDecode(f'No support for {ndim}-dimensional arrays')

        if memory_map:
            # offset calculations derived from numpy/core/memmap.py
            offset_header = file.tell()
            byte_count = offset_header + size * dtype.itemsize
            # ALLOCATIONGRANULARITY is 4096 on linux, if offset_header is 64 (or less than 4096), this will set offset_mmap to zero
            offset_mmap = offset_header - offset_header % mmap.ALLOCATIONGRANULARITY
            byte_count -= offset_mmap
            offset_array = offset_header - offset_mmap
            mm = mmap.mmap(file.fileno(),
                    byte_count,
                    access=mmap.ACCESS_READ,
                    offset=offset_mmap,
                    )
            # will always be immutable
            array = np.ndarray(shape,
                    dtype=dtype,
                    buffer=mm,
                    offset=offset_array,
                    order='F' if fortran_order else 'C',
                    )
            assert not array.flags.writeable
            return array, mm

        # NOTE: we cannot use np.from_file, as the file object from a Zip is not a normal file
        # NOTE: np.frombuffer produces a read-only view on the existing data
        # array = np.frombuffer(file.read(size * dtype.itemsize), dtype=dtype)
        array = np.frombuffer(file.read(), dtype=dtype)

        if fortran_order and ndim == 2:
            array.shape = (shape[1], shape[0])
            array = array.transpose()
        else:
            array.shape = shape
        assert not array.flags.writeable
        return array, None


class Archive:
    '''Abstraction of a directory or a zip archive.
    '''
    FILE_META = '__meta__.json'

    __slots__ = (
            'labels',
            'memory_map',
            '_archive',
            '_closable'
            )

    labels: tp.FrozenSet[str]
    memory_map: bool

    def __init__(self,
            fp: PathSpecifier,
            writeable: bool,
            memory_map: bool,
            ):
        raise NotImplementedError() #pragma: no cover

    def write_array(self, name: str, array: np.ndarray) -> None:
        raise NotImplementedError() #pragma: no cover

    def read_array(self, name: str) -> np.ndarray:
        raise NotImplementedError() #pragma: no cover

    def write_metadata(self, content: tp.Any) -> None:
        raise NotImplementedError() #pragma: no cover

    def read_metadata(self) -> tp.Any:
        raise NotImplementedError() #pragma: no cover

    def close(self) -> None:
        for f in getattr(self, '_closable', EMPTY_TUPLE):
            f.close()

class ArchiveZip(Archive):
    __slots__ = (
            'labels',
            'memory_map',
            '_archive',
            '_closable'
            )

    def __init__(self,
            fp: PathSpecifier,
            writeable: bool,
            memory_map: bool,
            ):

        mode = 'w' if writeable else 'r'
        self._archive = ZipFile(fp,
                mode=mode,
                compression=ZIP_STORED,
                allowZip64=True,
                )
        if not writeable:
            self.labels = frozenset(self._archive.namelist())
        if memory_map:
            raise RuntimeError(f'Cannot memory_map with {self}')

        self.memory_map = memory_map

    def __del__(self) -> None:
        self._archive.close()

    def write_array(self, name: str, array: np.ndarray) -> None:
        # NOTE: zip only was 'w' mode, not 'wb'
        # NOTE: force_zip64 required for large files; might make dynamic
        f = self._archive.open(name, 'w', force_zip64=True)
        try:
            NPYConverter.to_npy(f, array)
        finally:
            f.close()

    def read_array(self, name: str) -> np.ndarray:
        f = self._archive.open(name)
        try:
            array, _ = NPYConverter.from_npy(f)
        finally:
            f.close()
        array.flags.writeable = False
        return array

    def write_metadata(self, content: tp.Any) -> None:
        self._archive.writestr(self.FILE_META, json.dumps(content))

    def read_metadata(self) -> tp.Any:
        return json.loads(self._archive.read(self.FILE_META))


class ArchiveDirectory(Archive):
    __slots__ = (
            'labels',
            'memory_map',
            '_archive',
            '_closable',
            )

    def __init__(self,
            fp: PathSpecifier,
            writeable: bool,
            memory_map: bool,
            ):

        self._archive = fp
        if not os.path.exists(self._archive):
            if writeable:
                os.mkdir(fp)
            else:
                raise RuntimeError(f'Atttempting to read from a non-existant directory: {fp}')
        elif not os.path.isdir(self._archive):
            raise RuntimeError(f'A directory must be provided, not {fp}')

        if not writeable:
            self.labels = frozenset(f.name for f in os.scandir(self._archive))

        self.memory_map = memory_map

    def write_array(self, name: str, array: np.ndarray) -> None:
        fp = os.path.join(self._archive, name)
        f = open(fp, 'wb')
        try:
            NPYConverter.to_npy(f, array)
        finally:
            f.close()

    def read_array(self, name: str) -> np.ndarray:
        fp = os.path.join(self._archive, name)
        if self.memory_map:
            if not hasattr(self, '_closable'):
                self._closable = []

            f = open(fp, 'rb')
            try:
                array, mm = NPYConverter.from_npy(f, self.memory_map)
            finally:
                f.close() # NOTE: can close the file after creating memory map
            # self._closable.append(f)
            self._closable.append(mm)
            return array

        f = open(fp, 'rb')
        try:
            array, _ = NPYConverter.from_npy(f, self.memory_map)
        finally:
            f.close()
        return array

    def write_metadata(self, content: tp.Any) -> None:
        fp = os.path.join(self._archive, self.FILE_META)
        f = open(fp, 'w')
        try:
            f.write(json.dumps(content))
        finally:
            f.close()

    def read_metadata(self) -> tp.Any:
        fp = os.path.join(self._archive, self.FILE_META)
        f = open(fp, 'r')
        try:
            post = json.loads(f.read())
        finally:
            f.close()
        return post


class NPYArchiveConverter:
    KEY_NAMES = '__names__'
    KEY_TYPES = '__types__'
    KEY_DEPTHS = '__depths__'
    KEY_TYPES_INDEX = '__types_index__'
    KEY_TYPES_COLUMNS = '__types_columns__'
    FILE_TEMPLATE_VALUES_INDEX = '__values_index_{}__.npy'
    FILE_TEMPLATE_VALUES_COLUMNS = '__values_columns_{}__.npy'
    FILE_TEMPLATE_BLOCKS = '__blocks_{}__.npy'

    ARCHIVE_CLS: tp.Type[Archive]

    @staticmethod
    def _index_encode(
            *,
            metadata: tp.Dict[str, tp.Hashable],
            archive: Archive,
            index: 'IndexBase',
            key_template_values: str,
            key_types: str,
            depth: int,
            include: bool,
            ) -> None:
        '''
        Args:
            metadata: mutates in place with json components
        '''
        if depth == 1 and index._map is None: # type: ignore
            pass # do not store anything
        elif include:
            if depth == 1:
                archive.write_array(key_template_values.format(0), index.values)
            else:
                for i in range(depth):
                    archive.write_array(key_template_values.format(i), index.values_at_depth(i))
                metadata[key_types] = [cls.__name__ for cls in index.index_types.values] # type: ignore

    @classmethod
    def to_archive(cls,
            *,
            frame: 'Frame',
            fp: PathSpecifier,
            include_index: bool = True,
            include_columns: bool = True,
            ) -> None:
        '''
        Write a :obj:`Frame` as an npz file.
        '''
        metadata: tp.Dict[str, tp.Any] = {}
        metadata[cls.KEY_NAMES] = [frame._name,
                frame._index._name,
                frame._columns._name,
                ]
        # do not store Frame class as caller will determine
        metadata[cls.KEY_TYPES] = [
                frame._index.__class__.__name__,
                frame._columns.__class__.__name__,
                ]

        # store shape, index depths
        depth_index = frame._index.depth
        depth_columns = frame._columns.depth
        blocks = frame._blocks._blocks

        metadata[cls.KEY_DEPTHS] = [
                len(blocks),
                depth_index,
                depth_columns]

        archive = cls.ARCHIVE_CLS(fp,
                writeable=True,
                memory_map=False,
                )

        cls._index_encode(
                metadata=metadata,
                archive=archive,
                index=frame._index,
                key_template_values=cls.FILE_TEMPLATE_VALUES_INDEX,
                key_types=cls.KEY_TYPES_INDEX,
                depth=depth_index,
                include=include_index,
                )

        cls._index_encode(
                metadata=metadata,
                archive=archive,
                index=frame._columns,
                key_template_values=cls.FILE_TEMPLATE_VALUES_COLUMNS,
                key_types=cls.KEY_TYPES_COLUMNS,
                depth=depth_columns,
                include=include_columns,
                )

        for i, array in enumerate(blocks):
            archive.write_array(cls.FILE_TEMPLATE_BLOCKS.format(i), array)

        archive.write_metadata(metadata)

    @staticmethod
    def _index_decode(*,
            archive: Archive,
            metadata: tp.Dict[str, tp.Any],
            key_template_values: str,
            key_types: str,
            depth: int,
            cls_index: tp.Type['IndexBase'],
            name: NameType,
            ) -> tp.Optional['IndexBase']:
        '''Build index or columns.
        '''
        from static_frame.core.type_blocks import TypeBlocks

        if key_template_values.format(0) not in archive.labels:
            index = None
        elif depth == 1:
            index = cls_index(archive.read_array(key_template_values.format(0)),
                    name=name,
                    )
        else:
            index_tb = TypeBlocks.from_blocks(
                    archive.read_array(key_template_values.format(i))
                    for i in range(depth)
                    )
            index_constructors = [ContainerMap.str_to_cls(name)
                    for name in metadata[key_types]]
            index = cls_index._from_type_blocks(index_tb, # type: ignore
                    name=name,
                    index_constructors=index_constructors,
                    )
        return index

    @classmethod
    def from_archive(cls,
            *,
            constructor: tp.Type['Frame'],
            fp: PathSpecifier,
            memory_map: bool = False,
            ) -> 'Frame':
        '''
        Create a :obj:`Frame` from an npz file.
        '''
        from static_frame.core.type_blocks import TypeBlocks

        archive = cls.ARCHIVE_CLS(fp,
                writeable=False,
                memory_map=memory_map,
                )
        metadata = archive.read_metadata()

        # JSON will bring back tuple `name` attributes as lists; these must be converted to tuples to be hashable. Alternatives (like storing repr and using literal_eval) are slower than JSON.
        name, name_index, name_columns = (list_to_tuple(n) for n in metadata[cls.KEY_NAMES])

        block_count, depth_index, depth_columns = metadata[cls.KEY_DEPTHS]
        cls_index, cls_columns = (ContainerMap.str_to_cls(name)
                for name in metadata[cls.KEY_TYPES])

        index = cls._index_decode(
                archive=archive,
                metadata=metadata,
                key_template_values=cls.FILE_TEMPLATE_VALUES_INDEX,
                key_types=cls.KEY_TYPES_INDEX,
                depth=depth_index,
                cls_index=cls_index,
                name=name_index,
                )

        columns = cls._index_decode(
                archive=archive,
                metadata=metadata,
                key_template_values=cls.FILE_TEMPLATE_VALUES_COLUMNS,
                key_types=cls.KEY_TYPES_COLUMNS,
                depth=depth_columns,
                cls_index=cls_columns,
                name=name_columns,
                )

        tb = TypeBlocks.from_blocks(
                archive.read_array(cls.FILE_TEMPLATE_BLOCKS.format(i))
                for i in range(block_count)
                )

        if not memory_map:
            return constructor(tb,
                    own_data=True,
                    index=index,
                    own_index = False if index is None else True,
                    columns=columns,
                    own_columns = False if columns is None else True,
                    name=name,
                    )
        return constructor(tb,
                own_data=True,
                index=index,
                own_index = False if index is None else True,
                columns=columns,
                own_columns = False if columns is None else True,
                name=name,
                finalizer=archive.close,
                )


class NPZConverter(NPYArchiveConverter):
    ARCHIVE_CLS = ArchiveZip

class NPYDirectoryConverter(NPYArchiveConverter):
    ARCHIVE_CLS = ArchiveDirectory





import os
import unittest
from tempfile import TemporaryDirectory

import numpy as np
from numpy.lib.format import write_array # type: ignore

from static_frame.core.archive_npy import NPYConverter
from static_frame.core.archive_npy import ArchiveDirectory
from static_frame.core.archive_npy import ArchiveZip

from static_frame.core.exception import ErrorNPYDecode
from static_frame.core.exception import ErrorNPYEncode

from static_frame.test.test_case import temp_file
from static_frame.test.test_case import TestCase


class TestUnit(TestCase):


    #---------------------------------------------------------------------------
    def test_to_npy_a(self) -> None:
        a1 = np.arange(20)

        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                NPYConverter.to_npy(f, a1)

            a2 = np.load(fp)
            self.assertTrue((a1 == a2).all())

    def test_to_npy_b(self) -> None:
        a1 = np.array([None, 'foo', 3], dtype=object)

        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                with self.assertRaises(ErrorNPYEncode):
                    NPYConverter.to_npy(f, a1)

    def test_to_npy_c(self) -> None:
        a1 = np.arange(12).reshape(2, 3, 2)
        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                with self.assertRaises(ErrorNPYEncode):
                    NPYConverter.to_npy(f, a1)

    def test_to_npy_d(self) -> None:
        a1 = np.arange(12).reshape(2,6).T

        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                NPYConverter.to_npy(f, a1)
            with open(fp, 'rb') as f:
                a2, _ = NPYConverter.from_npy(f)

            self.assertTrue((a1 == a2).all())

    def test_to_npy_e(self) -> None:
        a1 = np.arange(4)
        with temp_file('.npy') as fp:

            with open(fp, 'wb') as f:
                NPYConverter.to_npy(f, a1)
            # ensure compatibility with numpy loaders
            a2 = np.load(fp)
            self.assertTrue((a1 == a2).all())

    def test_to_npy_f(self) -> None:
        a1 = np.array([('Rex', 9, 81.0), ('Fido', 3, 27.0)],
                dtype=[('name', 'U10'), ('age', 'i4'), ('weight', 'f4')]
                )
        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                with self.assertRaises(ErrorNPYEncode):
                    NPYConverter.to_npy(f, a1)




    def test_from_npy_a(self) -> None:
        a1 = np.arange(20)

        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                NPYConverter.to_npy(f, a1)
            with open(fp, 'rb') as f:
                a2, _ = NPYConverter.from_npy(f)
            self.assertTrue((a1 == a2).all())

    def test_from_npy_b(self) -> None:
        a1 = np.arange(100).reshape(5, 20)

        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                NPYConverter.to_npy(f, a1)
            with open(fp, 'rb') as f:
                a2, _ = NPYConverter.from_npy(f)

            self.assertTrue(a1.shape == a2.shape)
            self.assertTrue((a1 == a2).all())


    def test_from_npy_c(self) -> None:
        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                f.write(b'foo')

            with open(fp, 'rb') as f:
                # invaliud header raises
                with self.assertRaises(ErrorNPYDecode):
                    a2, _ = NPYConverter.from_npy(f)


    def test_from_npy_d(self) -> None:
        a1 = np.arange(12).reshape(2, 3, 2)

        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                write_array(f, a1, version=(1, 0))

            with open(fp, 'rb') as f:
                # invalid shape
                with self.assertRaises(ErrorNPYDecode):
                    a2, _ = NPYConverter.from_npy(f)

    def test_from_npy_e(self) -> None:
        a1 = np.array([2, 3, 4])

        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                write_array(f, a1, version=(3, 0))

            with open(fp, 'rb') as f:
                # invlid header; only version 1,0 is supported
                with self.assertRaises(ErrorNPYDecode):
                    a2, _ = NPYConverter.from_npy(f)


    def test_from_npy_f(self) -> None:
        a1 = np.array([None, 'foo', 3], dtype=object)

        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                write_array(f, a1, version=(1, 0))

            with open(fp, 'rb') as f:
                # invalid object dtype
                with self.assertRaises(ErrorNPYDecode):
                    a2, _ = NPYConverter.from_npy(f)


    def test_from_npy_g(self) -> None:
        a1 = np.array([2, 3, 4])

        with temp_file('.npy') as fp:
            with open(fp, 'wb') as f:
                NPYConverter.to_npy(f, a1)
            with open(fp, 'rb') as f:
                a2, _ = NPYConverter.from_npy(f, memory_map=True)
                self.assertEqual(a2.tolist(), [2, 3, 4])

    #---------------------------------------------------------------------------

    def test_archive_zip_a(self) -> None:
        with temp_file('.zip') as fp:
            with self.assertRaises(RuntimeError):
                _ = ArchiveZip(fp, writeable=True, memory_map=True)


    def test_archive_directory_a(self) -> None:
        with temp_file('.npy') as fp:
            with self.assertRaises(RuntimeError):
                ArchiveDirectory(fp, writeable=False, memory_map=False)

    def test_archive_directory_b(self) -> None:
        with TemporaryDirectory() as fp:
            os.rmdir(fp)
            # creates directory
            ad = ArchiveDirectory(fp, writeable=True, memory_map=False)

    def test_archive_directory_c(self) -> None:
        with TemporaryDirectory() as fp:
            os.rmdir(fp)
            # reading from a non-existant directory
            with self.assertRaises(RuntimeError):
                ad = ArchiveDirectory(fp, writeable=False, memory_map=False)
            os.mkdir(fp) # restore the directory for context manager

    def test_archive_directory_d(self) -> None:
        with TemporaryDirectory() as fp:
            a1 = np.arange(10)
            ad = ArchiveDirectory(fp, writeable=True, memory_map=False)
            ad.write_array('a1.npy', a1)
            a2 = ad.read_array('a1.npy')
            self.assertTrue((a1 == a2).all())

if __name__ == '__main__':
    unittest.main()




import typing as tp
from functools import partial

from static_frame.core.index import Index
from static_frame.core.index import IndexGO
from static_frame.core.util import PositionsAllocator
from static_frame.core.index_base import IndexBase  # pylint: disable = W0611
from static_frame.core.util import DTYPE_INT_DEFAULT
from static_frame.core.util import IndexConstructor
from static_frame.core.util import CallableOrMapping
from static_frame.core.util import IndexInitializer
from static_frame.core.util import NameType



class IndexDefaultFactory:
    '''
    Token class to be used to only provide a ``name`` to a default constructor of an Index.
    '''
    __slots__ = ('_name',)

    def __init__(self, name: NameType):
        self._name = name

    def __call__(self, constructor: IndexConstructor) -> IndexConstructor:
        '''Partial the passeed constructor with the ``name``.
        '''
        return partial(constructor, name=self._name)

IndexAutoInitializer = int

# could create trival subclasses for these indices, but the type would would not always describe the instance; for example, an IndexAutoGO could grow inot non-contiguous integer index, as loc_is_iloc is reevaluated with each append can simply go to false.

class IndexAutoFactory:
    __slots__ = ('_size', '_name')

    @classmethod
    def from_optional_constructor(cls,
            initializer: IndexAutoInitializer, # size
            *,
            default_constructor: tp.Type['IndexBase'],
            explicit_constructor: tp.Optional[tp.Union[IndexConstructor, IndexDefaultFactory]] = None,
            ) -> 'IndexBase':

        # get an immutable array, shared from positions allocator
        labels = PositionsAllocator.get(initializer)

        if explicit_constructor:
            if isinstance(explicit_constructor, IndexDefaultFactory):
                # partial the default constructor with a name argument
                return explicit_constructor(default_constructor)(labels)
            return explicit_constructor(labels)

        else: # get from default constructor
            constructor = Index if default_constructor.STATIC else IndexGO
            return constructor(
                    labels=labels,
                    loc_is_iloc=True,
                    dtype=DTYPE_INT_DEFAULT
                    )

    def __init__(self,
            size: IndexAutoInitializer,
            *,
            name: NameType = None,
            ):
        self._size = size
        self._name = name

    def to_index(self,
            *,
            default_constructor: tp.Type['IndexBase'],
            explicit_constructor: tp.Optional[tp.Union[IndexConstructor, IndexDefaultFactory]] = None,
            ) -> 'IndexBase':
        '''Called by index_from_optional_constructor.
        '''
        return self.from_optional_constructor(self._size,
                default_constructor=default_constructor,
                explicit_constructor=explicit_constructor,
                )



IndexAutoFactoryType = tp.Type[IndexAutoFactory]
RelabelInput = tp.Union[CallableOrMapping, IndexAutoFactoryType, IndexInitializer]



import sys
import os
import typing as tp

import invoke

#-------------------------------------------------------------------------------

@invoke.task
def clean(context):
    '''Clean doc and build artifacts
    '''
    context.run('rm -rf coverage.xml')
    context.run('rm -rf htmlcov')
    context.run('rm -rf doc/build')
    context.run('rm -rf build')
    context.run('rm -rf dist')
    context.run('rm -rf *.egg-info')
    context.run('rm -rf .coverage')
    context.run('rm -rf .mypy_cache')
    context.run('rm -rf .pytest_cache')
    context.run('rm -rf .hypothesis')
    context.run('rm -rf .ipynb_checkpoints')



@invoke.task()
def doc(context):
    '''Build docs
    '''
    context.run(f'{sys.executable} doc/doc_build.py')


@invoke.task
def performance(context):
    '''Run performance tests.
    '''
    # NOTE: we do not get to see incremental output when running this
    cmd = 'python static_frame/performance/main.py --performance "*"'
    context.run(cmd)


@invoke.task
def interface(context, container=None):
    '''
    Optionally select a container type to discover what API endpoints have examples.
    '''
    from static_frame.core.container import ContainerBase
    import static_frame as sf


    def subclasses(cls) -> tp.Iterator[tp.Type]:
        if cls.__name__ not in ('IndexBase', 'IndexDatetime'):
            yield cls
        for sub in cls.__subclasses__():
            yield from subclasses(sub)

    if not container:
        def frames():
            for cls in sorted(subclasses(ContainerBase),
                    key=lambda cls: cls.__name__):
                yield cls.interface.unset_index()
        f = sf.Frame.from_concat(frames(), axis=0, index=sf.IndexAutoFactory)
    else:
        f = getattr(sf, container).interface

    print(f.display_tall())

@invoke.task
def example(context, container=None):
    '''
    Discover API members that have a code example.
    '''
    from static_frame.core.display_color import HexColor
    from doc.source.conf import get_jinja_contexts

    contexts = get_jinja_contexts()
    defined = contexts['examples_defined']
    signatures = set()

    # discover all signatures; if it is defined, print in a darker color
    for name, cls, frame in contexts['interface'].values():
        for signature, row in frame.iter_tuple_items(axis=1):
            target = f'{name}-{row.signature_no_args}'
            signatures.add(target) # accumulate all signatures
            if container and name != container:
                continue
            if target in defined:
                print(HexColor.format_terminal(0x505050, target))
            else:
                print(target)

    for line in sorted(defined - signatures):
        print(HexColor.format_terminal(0x00ccff, line))


#-------------------------------------------------------------------------------

@invoke.task
def test(context,
        unit=False,
        cov=False,
        ):
    '''Run tests.
    '''
    if unit:
        fp = 'static_frame/test/unit'
    else:
        fp = 'static_frame/test'

    cmd = f'pytest -s --color no --disable-pytest-warnings --tb=native {fp}'

    if cov:
        cmd += ' --cov=static_frame --cov-report=xml'

    print(cmd)
    context.run(cmd)


@invoke.task
def coverage(context):
    '''
    Perform code coverage, and open report HTML.
    '''
    cmd = 'pytest -s --color no --disable-pytest-warnings --cov=static_frame/core --cov-report html'
    print(cmd)
    context.run(cmd)
    import webbrowser
    webbrowser.open('htmlcov/index.html')


@invoke.task
def mypy(context):
    '''Run mypy static analysis.
    '''
    context.run('mypy --strict')


@invoke.task
def lint(context):
    '''Run pylint static analysis.
    '''
    context.run('pylint -f colorized static_frame')

@invoke.task(pre=(mypy, lint))
def quality(context):
    '''Perform all quality checks.
    '''

#-------------------------------------------------------------------------------

@invoke.task(pre=(clean,))
def build(context):
    '''Build packages
    '''
    context.run(f'{sys.executable} setup.py sdist bdist_wheel')

@invoke.task(pre=(build,), post=(clean,))
def release(context):
    context.run('twine upload dist/*')





import typing as tp
import time
import sys
import traceback
import random
import argparse
import subprocess

from static_frame.core.display_color import HexColor


class Line:
    pass

PAUSE_SHORT = Line()
PAUSE_LONG = Line()
PAUSE_FINAL = Line()

class Comment(Line):
    def __init__(self, message: str, color: int = 0xaaaaaa) -> None:
        self.message = message
        self.color = color

    def __iter__(self) -> tp.Iterator[str]:
        return HexColor.format_terminal(self.color, self.message).__iter__()

LineIter = tp.Iterator[tp.Union[Line, str]]

#-------------------------------------------------------------------------------
class LineGen:
    CMD_PREFIX = ''

    @staticmethod
    def lines() -> LineIter:
        raise NotImplementedError()

    @classmethod
    def get_animate_command(cls) -> str:
        if cls.CMD_PREFIX:
            return f"{cls.CMD_PREFIX} python3 doc/animate/animator.py --animate {cls.__name__}"
        return f"python3 doc/animate/animator.py --animate {cls.__name__}"



class DisplayConfig(LineGen):

    @staticmethod
    def lines() -> LineIter:
        yield 'import static_frame as sf'
        yield PAUSE_SHORT

        yield "f = sf.FrameGO.from_dict_records_items((('charm', {'symbol':'c', 'mass':1.3}), ('strange', {'symbol':'s', 'mass':0.1})))"
        yield "f['target'] = f['mass'] > 1"
        yield 'f'
        yield PAUSE_SHORT


        yield "sf.DisplayActive.set(sf.DisplayConfig(type_color_float='orange'))"
        yield 'f'
        yield PAUSE_SHORT

        yield "sf.DisplayActive.update(type_color_str='lightcoral', type_color_bool='mediumpurple')"
        yield 'f'
        yield PAUSE_SHORT


        yield "sf.DisplayActive.update(type_color_index=0x4682b4, type_color_frame=0xbfff)"
        yield 'f'
        yield PAUSE_SHORT

        yield 'f.display(sf.DisplayConfig(type_color=False))'
        yield PAUSE_SHORT

        yield 'f.display(sf.DisplayConfig(type_show=False, include_index=False))'
        yield PAUSE_FINAL


class LowMemoryOpsVerbose(LineGen):
    CMD_PREFIX = 'prlimit --as=800000000' # shown to cause expected memory error

    @staticmethod
    def lines() -> LineIter:
        # return lines of code to execute

        yield Comment("# This example demonstrates one of the many benefits of StaticFrame's use of immutable data by simulating a low-memory environment with prlimit. Let us start by importing numpy, pandas, and static_frame")
        yield PAUSE_SHORT

        yield 'import numpy as np'
        yield 'import pandas as pd'
        yield 'import static_frame as sf'
        yield PAUSE_SHORT

        yield Comment('# We will create a large 2D array of integers and a tuple of column labels.')
        yield PAUSE_SHORT

        yield 'a1 = np.arange(10_000_000).reshape(1_000_000, 10)'
        yield "columns = tuple('abcdefghij')"
        yield PAUSE_SHORT

        yield Comment('# Next, we create a Pandas DataFrame using that array.')
        yield PAUSE_SHORT


        yield 'df1 = pd.DataFrame(a1, columns=columns)'
        yield 'df1.shape'
        yield PAUSE_LONG


        yield Comment('# Pandas cannot rename the DataFrame without defensively copying the data, which in this low-memory environment causes a MemoryError.')
        yield PAUSE_SHORT

        yield 'df1.rename(columns=lambda x: x.upper())'
        yield PAUSE_LONG


        yield Comment('# Similarly, concatenating the DataFrame with itself results in a MemoryError.')
        yield PAUSE_SHORT

        yield 'pd.concat((df1, df1), axis=1, ignore_index=True)'
        yield PAUSE_LONG


        yield Comment('# To reuse the same array in StaticFrame, we can make it immutable.')
        yield PAUSE_SHORT

        yield 'a1.flags.writeable = False'
        yield PAUSE_SHORT

        yield 'f1 = sf.Frame(a1, columns=columns)'
        yield 'f1.shape'
        yield PAUSE_SHORT


        yield Comment('# As StaticFrame is built on immutable arrays, we can relabel the Frame without a MemoryError, as underlying data does not need to be copied.')
        yield PAUSE_SHORT

        yield 'f2 = f1.relabel(columns=lambda x: x.upper())'
        yield 'f2.shape'
        yield 'f2.columns'
        yield PAUSE_LONG


        yield Comment('# Similarly, while Pandas runs out of memory, StaticFrame can successfully concatenate the Frame.')
        yield PAUSE_SHORT

        yield 'f3 = sf.Frame.from_concat((f1, f2), axis=1)'
        yield 'f3.columns.values'
        yield 'f3.shape'
        yield PAUSE_LONG


class LowMemoryQuilt(LineGen):
    CMD_PREFIX = 'prlimit --as=800000000' # shown to cause expected memory error

    @staticmethod
    def lines() -> LineIter:
        yield 'import numpy as np'
        yield 'import pandas as pd'
        yield 'import static_frame as sf'
        yield 'from string import ascii_lowercase'
        # yield "df = pd.DataFrame(np.arange(6_000_000 * 4).reshape(6_000_000, 4), columns=tuple('wxyz'))"
        # yield "del df"
        yield "fp = '/tmp/test.zip'"
        yield 'config = sf.StoreConfig(include_index=True, index_depth=1)'
        yield "items = ((ascii_lowercase[i], sf.Frame(np.arange(2_000_000).reshape(500_000, 4), columns=tuple('wxyz'))) for i in range(12))"
        yield 'sf.Batch(items).to_zip_pickle(fp, config=config)'
        yield 'q1 = sf.Quilt.from_zip_pickle(fp, max_persist=1, retain_labels=True, config=config, deepcopy_from_bus=True)'
        yield 'q1.shape'
        yield "q1.iloc[1000000:1000010, 3:]"
        # yield 'sf.Batch(q1.iter_window_items(size=100_000, step=100_000)).mean().to_frame()'



#------------------------------------------------------------------------\------
class Animator:

    PROMPT = HexColor.format_terminal('lightgrey', '>>> ')
    CHAR_INTERVAL = 0.03 #0.07
    CHAR_JITTER = [x * .01 for x in range(6)] + [0.10, .12]

    @classmethod
    def print_char(cls, char: str) -> None:
        print(char, end='')
        sys.stdout.flush()
        # time.sleep(cls.CHAR_INTERVAL + random.choice(cls.CHAR_JITTER))

    @classmethod
    def pause(cls, interval: float) -> None:
        print(cls.PROMPT, end='')
        sys.stdout.flush()
        time.sleep(interval)
        print() # newline
        sys.stdout.flush()


    @classmethod
    def main(cls, func: tp.Callable[[], LineIter]) -> None:

        import static_frame as sf
        import numpy as np
        from string import ascii_lowercase

        for line in func():
            if line is PAUSE_SHORT:
                cls.pause(1)
                continue
            if line is PAUSE_LONG:
                cls.pause(2.5)
                continue
            if line is PAUSE_FINAL:
                cls.pause(5)
                continue

            assert isinstance(line, (Comment, str))

            print(cls.PROMPT, end='')
            for char in line:
                cls.print_char(char)
            cls.print_char('\n') # get new line

            if isinstance(line, Comment):
                continue

            # NOTE: exec puts variables in this scope; pass globals and locals eval, exec
            g = globals()
            g['sf'] = sf
            g['np'] = np
            g['ascii_lowercase'] = ascii_lowercase
            l = locals()
            try:
                post = eval(line, g, l)
                if post is not None:
                    print(post)
            except SyntaxError:
                exec(line, g, l)
            except MemoryError as e:
                traceback.print_exc(limit=-3)


def get_arg_parser() -> argparse.ArgumentParser:
    p = argparse.ArgumentParser(
            description='Terminal animator',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            )
    # NOTE: when using --animate, CMD_PREFIX must be applied when the command is called
    p.add_argument('--animate',
            help='Name of class to display the animation.',
            )
    p.add_argument('--termtosvg',
            help='Name of class to termtosvg.',
            )
    p.add_argument('--asciinema',
            help='Name of class to asciinema.',
            )
    return p


if __name__ == '__main__':

    options = get_arg_parser().parse_args()
    line_gen = {cls.__name__: cls for cls in (
            LowMemoryOpsVerbose,
            DisplayConfig,
            LowMemoryQuilt,
            )}

    if options.animate:
        cls = line_gen[options.animate]
        Animator.main(cls.lines)

    elif options.termtosvg:
        cls = line_gen[options.termtosvg]

        cmd = ['termtosvg',
            '--template',
            'window_frame',
            '-g', '80x18',
            '--loop-delay', '2000',
            '--command',
            cls.get_animate_command(),
            '/tmp/term.svg',
            ]
        subprocess.run(cmd)

    elif options.asciinema:
        cls = line_gen[options.asciinema]

        cmd = ['asciinema',
            'rec',
            '--command',
            cls.get_animate_command(),
            '/tmp/term.cast',
            ]
        subprocess.run(cmd)
        # NOTE: can upload with asciinema upload /tmp/term.cast




import os
import timeit
import tempfile
import typing as tp
import pickle
import shutil

import matplotlib.pyplot as plt
import numpy as np
import frame_fixtures as ff
import static_frame as sf
import pandas as pd
from static_frame.core.display_color import HexColor



FF_wide = 's(10,10_000)|v(int,int,bool,float,float)|i(I,int)|c(I,str)'
FF_wide_col = 's(10,10_000)|v(int,bool,float)|i(I,int)|c(I,str)'
FF_wide_ext = 's(1000,10_000)|v(int,int,bool,float,float)|i(I,int)|c(I,str)'

FF_tall = 's(10_000,10)|v(int,int,bool,float,float)|i(I,int)|c(I,str)'
FF_tall_col = 's(10_000,10)|v(int,bool,float)|i(I,int)|c(I,str)'
FF_tall_ext = 's(10_000,1000)|v(int,int,bool,float,float)|i(I,int)|c(I,str)'

FF_square = 's(1_000,1_000)|v(float)|i(I,int)|c(I,str)'

class FileIOTest:
    NUMBER = 4
    SUFFIX = '.tmp'

    def __init__(self, fixture: str):
        self.fixture = ff.parse(fixture)
        _, self.fp = tempfile.mkstemp(suffix=self.SUFFIX)
        self.fp_dir = '/tmp/npy'

    def clear(self) -> None:
        os.unlink(self.fp)
        if os.path.exists(self.fp_dir):
            shutil.rmtree(self.fp_dir)

    def __call__(self):
        raise NotImplementedError()



class SFReadParquet(FileIOTest):
    SUFFIX = '.parquet'

    def __init__(self, fixture: str):
        super().__init__(fixture)
        self.fixture.to_parquet(self.fp)

    def __call__(self):
        f = sf.Frame.from_parquet(self.fp, index_depth=1)
        _ = f.loc[34715, 'zZbu']

class SFWriteParquet(FileIOTest):
    SUFFIX = '.parquet'

    def __call__(self):
        self.fixture.to_parquet(self.fp)



class PDReadParquetArrow(FileIOTest):
    SUFFIX = '.parquet'

    def __init__(self, fixture: str):
        super().__init__(fixture)
        df = self.fixture.to_pandas()
        df.to_parquet(self.fp)

    def __call__(self):
        f = pd.read_parquet(self.fp)
        _ = f.loc[34715, 'zZbu']

class PDWriteParquetArrow(FileIOTest):
    SUFFIX = '.parquet'

    def __init__(self, fixture: str):
        super().__init__(fixture)
        self.df = self.fixture.to_pandas()

    def __call__(self):
        self.df.to_parquet(self.fp)


class PDReadParquetFast(FileIOTest):
    SUFFIX = '.parquet'

    def __init__(self, fixture: str):
        super().__init__(fixture)
        df = self.fixture.to_pandas()
        df.to_parquet(self.fp, engine='fastparquet')

    def __call__(self):
        f = pd.read_parquet(self.fp, engine='fastparquet')
        _ = f.loc[34715, 'zZbu']

class PDWriteParquetFast(FileIOTest):
    SUFFIX = '.parquet'

    def __init__(self, fixture: str):
        super().__init__(fixture)
        self.df = self.fixture.to_pandas()

    def __call__(self):
        self.df.to_parquet(self.fp, engine='fastparquet')





class SFReadNPZ(FileIOTest):
    SUFFIX = '.npz'

    def __init__(self, fixture: str):
        super().__init__(fixture)
        self.fixture.to_npz(self.fp)

    def __call__(self):
        f = sf.Frame.from_npz(self.fp)
        _ = f.loc[34715, 'zZbu']


class SFWriteNPZ(FileIOTest):
    SUFFIX = '.npz'

    def __call__(self):
        self.fixture.to_npz(self.fp)



class SFReadPickle(FileIOTest):
    SUFFIX = '.pickle'

    def __init__(self, fixture):
        super().__init__(fixture)
        self.file = open(self.fp, 'wb')
        pickle.dump(self.fixture, self.file)
        self.file.close()

    def __call__(self):
        with open(self.fp, 'rb') as f:
            f = pickle.load(f)
            _ = f.loc[34715, 'zZbu']


    def clear(self) -> None:
        self.file.close()
        os.unlink(self.fp)


class SFWritePickle(FileIOTest):
    SUFFIX = '.pickle'

    def __init__(self, fixture):
        super().__init__(fixture)
        self.file = open(self.fp, 'wb')

    def __call__(self):
        pickle.dump(self.fixture, self.file)

    def clear(self) -> None:
        self.file.close()
        os.unlink(self.fp)



class SFReadNPY(FileIOTest):
    SUFFIX = '.npy'

    def __init__(self, fixture: str):
        super().__init__(fixture)
        self.fixture.to_npy(self.fp_dir)

    def __call__(self):
        # import ipdb; ipdb.set_trace()
        f = sf.Frame.from_npy(self.fp_dir)
        _ = f.loc[34715, 'zZbu']


class SFWriteNPY(FileIOTest):
    SUFFIX = '.npy'

    def __call__(self):
        self.fixture.to_npy(self.fp_dir)




class SFReadNPYMM(FileIOTest):
    SUFFIX = '.npy'

    def __init__(self, fixture: str):
        super().__init__(fixture)
        self.fixture.to_npy(self.fp_dir)

    def __call__(self):
        # import ipdb; ipdb.set_trace()
        f = sf.Frame.from_npy(self.fp_dir, memory_map=True)
        _ = f.loc[34715, 'zZbu']



def plot(frame: sf.Frame):
    from collections import defaultdict
    fig, axes = plt.subplots(2, 1)

    for axes_count, cat in enumerate(frame.iter_group('category')):
        ax = axes[axes_count]

        labels = [] # by fixture
        results = defaultdict(list)

        components = 0
        for label, sub in cat.iter_group_items('fixture'):
            labels.append(label)
            components = max(components, len(sub))
            for row in sub.iter_series(axis=1):
                results[row['name']].append(row['time'])

        width = 0.85  # the width of each group
        x = np.arange(len(labels))

        segment = width / components
        start = -width / 2
        for i, (label, values) in enumerate(results.items()):
            r = ax.bar(x + (start + (segment * i)), values, segment, label=label)
            # ax.bar_label(r, padding=3)

        # Add some text for labels, title and custom x-axis tick labels, etc.
        # ax.set_ylabel('Scores')
        # ax.set_title('Scores by group and gender')
        ax.set_xticks(x, labels)
        ax.set_yscale('log')
        ax.legend()


    fig.tight_layout()
    fp = '/tmp/serialize.png'
    plt.savefig(fp, dpi=300)
    import os
    os.system(f'eog {fp}')


#-------------------------------------------------------------------------------
def convert_size(size_bytes):
    import math
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])

def get_sizes():
    records = []
    for label, fixture in (
            ('wide', FF_wide),
            ('wide_ext', FF_wide_ext),
            ('tall', FF_tall),
            ('tall_ext', FF_tall_ext),
            ('square', FF_square),
            ):
        f = ff.parse(fixture)
        record = [label, f.shape]

        _, fp = tempfile.mkstemp(suffix='.parquet')
        f.to_parquet(fp, include_index=True)
        size = os.path.getsize(fp)
        os.unlink(fp)
        record.append(convert_size(size))

        _, fp = tempfile.mkstemp(suffix='.npz')
        f.to_npz(fp, include_columns=True)
        size = os.path.getsize(fp)
        os.unlink(fp)
        record.append(convert_size(size))

        _, fp = tempfile.mkstemp(suffix='.pickle')
        file = open(fp, 'wb')
        pickle.dump(f, file)
        file.close()
        size = os.path.getsize(fp)
        os.unlink(fp)
        record.append(convert_size(size))

        records.append(record)

    f = sf.Frame.from_records(records, columns=('fixture', 'shape', 'parquet', 'npz', 'pickle')).set_index('fixture', drop=True)
    print(f)


def pandas_serialize():
    import pandas as pd
    df = ff.parse('s(10,10)|v(int,int,bool,float,float)|i(I,int)|c(I,str)').rename('foo').to_pandas().set_index(['zZbu', 'ztsv'])
    df = df.reindex(columns=pd.Index(df.columns, name='foo'))



# This subclass of ndarray has some unpleasant interactions with some operations, because it doesn’t quite fit properly as a subclass. An alternative to using this subclass is to create the mmap object yourself, then create an ndarray with ndarray.__new__ directly, passing the object created in its ‘buffer=’ parameter.

#-------------------------------------------------------------------------------


def get_format():

    name_root_last = None
    name_root_count = 0

    def format(key: tp.Tuple[tp.Any, str], v: object) -> str:
        nonlocal name_root_last
        nonlocal name_root_count

        if isinstance(v, float):
            if np.isnan(v):
                return ''
            return str(round(v, 4))
        if isinstance(v, (bool, np.bool_)):
            if v:
                return HexColor.format_terminal('green', str(v))
            return HexColor.format_terminal('orange', str(v))

        return str(v)

    return format

from itertools import repeat
from itertools import chain

def run_test():
    records = []
    for label, fixture in (
            ('wide', FF_wide),
            ('wide_col', FF_wide_col),
            ('tall', FF_tall),
            ('tall_col', FF_tall_col),
            ('square', FF_square),
            ):
        cls_read = (
            PDReadParquetArrow,
            # PDReadParquetFast, # not faster!
            SFReadParquet,
            SFReadNPZ,
            SFReadNPY,
            SFReadPickle,
            # SFReadNPYMM,
            )
        cls_write = (
            PDWriteParquetArrow,
            # PDWriteParquetFast, # not faster!
            SFWriteParquet,
            SFWriteNPZ,
            SFWriteNPY,
            SFWritePickle,
            )

        for cls, category in chain(
                zip(cls_read, repeat('read')),
                zip(cls_write, repeat('write')),
                ):
            runner = cls(fixture)
            record = [cls.__name__, cls.NUMBER, category, label]
            try:
                result = timeit.timeit(
                        f'runner()',
                        globals=locals(),
                        number=cls.NUMBER)
            except OSError:
                result = np.nan
            finally:
                runner.clear()
            record.append(result)
            records.append(record)

    f = sf.FrameGO.from_records(records,
            columns=('name', 'number', 'category', 'fixture', 'time')
            )

    display = f.iter_element_items().apply(get_format())

    config = sf.DisplayConfig(
            cell_max_width_leftmost=np.inf,
            cell_max_width=np.inf,
            type_show=False,
            display_rows=200,
            include_index=False,
            )
    print(display.display(config))

    plot(f)
    # import ipdb; ipdb.set_trace()

if __name__ == '__main__':
    # pandas_serialize()
    # get_sizes()
    run_test()





import unittest

import numpy as np

from static_frame.test.test_case import TestCase
from static_frame.core.rank import rank_1d
from static_frame.core.rank import rank_2d

from static_frame.core.rank import RankMethod

class TestUnit(TestCase):

    def test_rank_method_a(self) -> None:
        with self.assertRaises(NotImplementedError):
            rank_1d(np.array([3, 2, 6, 20]), None) #type: ignore

    def test_rank_ordinal_a(self) -> None:
        self.assertEqual(
                rank_1d(np.array([3, 2, 6, 20]), RankMethod.ORDINAL).tolist(),
                [1, 0, 2, 3]
                )

        self.assertEqual(
                rank_1d(np.array([3, 2, 6, 20]), RankMethod.ORDINAL, start=1).tolist(),
                [2, 1, 3, 4]
                )


        a2 = rank_1d(np.array([3, 2, 6, 20]), RankMethod.ORDINAL, ascending=False)
        self.assertEqual(a2.tolist(),
                [2, 3, 1, 0]
                )

    def test_rank_ordinal_b(self) -> None:

        a1 = rank_1d(np.array([0, 2, 3, 2]), 'ordinal', start=1)
        self.assertEqual(a1.tolist(),
                [1, 2, 4, 3]
                )

        a2 = rank_1d(np.array([0, 2, 3, 2]), 'ordinal', start=0)
        self.assertEqual(a2.tolist(),
                [0, 1, 3, 2]
                )

        a3 = rank_1d(np.array([0, 2, 3, 2]), 'ordinal', ascending=False)
        self.assertEqual(a3.tolist(),
                [3, 2, 0, 1]
                )

    def test_rank_ordinal_c(self) -> None:
        a1 = rank_1d(np.array([8, 15, 7, 2, 20, 4, 20, 7, 15, 15]), 'ordinal', start=1)
        self.assertEqual(a1.tolist(),
                [5, 6, 3, 1, 9, 2, 10, 4, 7, 8]
                )
        #scipy: [5, 6, 3, 1, 9, 2, 10, 4, 7, 8]

    def test_rank_ordinal_d(self) -> None:
        a1 = rank_1d(np.array([17, 10, 3, -4, 9, 3, -12, 18, 3, 0, 17]), 'ordinal', start=1)
        self.assertEqual(a1.tolist(),
                [9, 8, 4, 2, 7, 5, 1, 11, 6, 3, 10]
                )
        #scipy: [9, 8, 4, 2, 7, 5, 1, 11, 6, 3, 10]







    def test_rank_average_a(self) -> None:

        a1 = rank_1d(np.array([0, 2, 3, 2]), 'mean', ascending=True)
        self.assertEqual(a1.tolist(),
                [0.0, 1.5, 3.0, 1.5]
                )

        a2 = rank_1d(np.array([0, 2, 3, 2]), 'mean', start=1)
        self.assertEqual(a2.tolist(),
                [1.0, 2.5, 4.0, 2.5]
                )

        a3 = rank_1d(np.array([0, 2, 3, 2]), 'mean', start=1, ascending=False)
        self.assertEqual(a3.tolist(),
                [4.0, 2.5, 1.0, 2.5]
                )


    def test_rank_average_b(self) -> None:

        a1 = rank_1d(np.array([0, 2, 5, 2, 2, 2]), 'mean', ascending=True)
        self.assertEqual(a1.tolist(),
                [0.0, 2.5, 5.0, 2.5, 2.5, 2.5]
                )

        a1 = rank_1d(np.array([0, 2, 5, 2, 2, 2]), 'mean', ascending=True, start=1)
        self.assertEqual(a1.tolist(),
                [1.0, 3.5, 6.0, 3.5, 3.5, 3.5]
                )
        #scipy: [1.0, 3.5, 6.0, 3.5, 3.5, 3.5]

        # import ipdb; ipdb.set_trace()
        a2 = rank_1d(np.array([0, 2, 3, 2]), 'mean', start=1)
        self.assertEqual(a2.tolist(),
                [1.0, 2.5, 4.0, 2.5]
                )

        a3 = rank_1d(np.array([0, 2, 3, 2]), 'mean', start=1, ascending=False)
        self.assertEqual(a3.tolist(),
                [4.0, 2.5, 1.0, 2.5]
                )

    def test_rank_average_c(self) -> None:
        a1 = rank_1d(np.array([8, 15, 7, 2, 20, 4, 20, 7, 15, 15]), 'mean', start=1)
        self.assertEqual(a1.tolist(),
                [5.0, 7.0, 3.5, 1.0, 9.5, 2.0, 9.5, 3.5, 7.0, 7.0]
                )
        #scipy: [5.0, 7.0, 3.5, 1.0, 9.5, 2.0, 9.5, 3.5, 7.0, 7.0]

    def test_rank_average_d(self) -> None:
        a1 = rank_1d(np.array([17, 10, 3, -4, 9, 3, -12, 18, 3, 0, 17]), 'mean', start=1)
        self.assertEqual(a1.tolist(),
                [9.5, 8.0, 5.0, 2.0, 7.0, 5.0, 1.0, 11.0, 5.0, 3.0, 9.5]
                )
        #scipy: [9.5, 8.0, 5.0, 2.0, 7.0, 5.0, 1.0, 11.0, 5.0, 3.0, 9.5






    def test_rank_min_a(self) -> None:

        a1 = rank_1d(np.array([0, 2, 3, 2]), 'min', start=1)
        self.assertEqual(a1.tolist(),
                [1, 2, 4, 2]
                )
        a2 = rank_1d(np.array([0, 2, 3, 2]), 'min', start=0)
        self.assertEqual(a2.tolist(),
                [0, 1, 3, 1]
                )

        a3 = rank_1d(np.array([0, 2, 3, 2]), 'min', ascending=False)
        self.assertEqual(a3.tolist(),
                [3, 1, 0, 1]
                )

    def test_rank_min_b(self) -> None:
        a1 = rank_1d(np.array([8, 15, 7, 2, 20, 4, 20, 7, 15, 15]), 'min', start=1)
        self.assertEqual(a1.tolist(),
                [5, 6, 3, 1, 9, 2, 9, 3, 6, 6]
                )
        #scipy: [5, 6, 3, 1, 9, 2, 9, 3, 6, 6]

    def test_rank_min_c(self) -> None:
        a1 = rank_1d(np.array([17, 10, 3, -4, 9, 3, -12, 18, 3, 0, 17]), 'min', start=1)
        self.assertEqual(a1.tolist(),
                [9, 8, 4, 2, 7, 4, 1, 11, 4, 3, 9]
                )

        #scipy: [9, 8, 4, 2, 7, 4, 1, 11, 4, 3, 9]






    def test_rank_max_a(self) -> None:

        a1 = rank_1d(np.array([0, 2, 3, 2]), 'max', start=1)
        # import ipdb; ipdb.set_trace()

        self.assertEqual(a1.tolist(),
                [1, 3, 4, 3]
                )
        a2 = rank_1d(np.array([0, 2, 3, 2]), 'max', start=0)
        self.assertEqual(a2.tolist(),
                [0, 2, 3, 2]
                )

        a2 = rank_1d(np.array([0, 2, 3, 2]), 'max', ascending=False)
        self.assertEqual(a2.tolist(),
                [3, 2, 0, 2]
                )

    def test_rank_max_b(self) -> None:
        a1 = rank_1d(np.array([8, 15, 7, 2, 20, 4, 20, 7, 15, 15]), 'max', start=1)
        self.assertEqual(a1.tolist(),
                [5, 8, 4, 1, 10, 2, 10, 4, 8, 8]
                )
        #scipy: [5, 8, 4, 1, 10, 2, 10, 4, 8, 8]

    def test_rank_max_c(self) -> None:
        a1 = rank_1d(np.array([17, 10, 3, -4, 9, 3, -12, 18, 3, 0, 17]), 'max', start=1)
        self.assertEqual(a1.tolist(),
                [10, 8, 6, 2, 7, 6, 1, 11, 6, 3, 10]
                )
        #scipy: [10, 8, 6, 2, 7, 6, 1, 11, 6, 3, 10]




    def test_rank_dense_a(self) -> None:

        a1 = rank_1d(np.array([0, 2, 3, 2]), 'dense', start=1)
        self.assertEqual(a1.tolist(),
                [1, 2, 3, 2]
                )

        a2 = rank_1d(np.array([0, 2, 3, 2]), 'dense', start=0)
        self.assertEqual(a2.tolist(),
                [0, 1, 2, 1]
                )

        a3 = rank_1d(np.array([0, 2, 3, 2]), 'dense', ascending=False)
        self.assertEqual(a3.tolist(),
                [2, 1, 0, 1]
                )

    def test_rank_dense_b(self) -> None:
        a1 = rank_1d(np.array([8, 15, 7, 2, 20, 4, 20, 7, 15, 15]), 'dense', start=1)
        self.assertEqual(a1.tolist(),
                [4, 5, 3, 1, 6, 2, 6, 3, 5, 5]
                )
        #scipy: [4, 5, 3, 1, 6, 2, 6, 3, 5, 5]

    def test_rank_dense_c(self) -> None:
        a1 = rank_1d(np.array([17, 10, 3, -4, 9, 3, -12, 18, 3, 0, 17]), 'dense', start=1)
        self.assertEqual(a1.tolist(),
                [7, 6, 4, 2, 5, 4, 1, 8, 4, 3, 7]
                )
        #scipy: [7, 6, 4, 2, 5, 4, 1, 8, 4, 3, 7]





    def test_rank_2d_a(self) -> None:
        a1 = np.array([10, 3, -4, 9, 3, -12, 18, 3, 0, 17]).reshape(5,2)
        self.assertEqual(
            rank_2d(a1, axis=0, method='ordinal', start=1).tolist(),
            [[4, 2], [1, 4], [3, 1], [5, 3], [2, 5]]
        )
        self.assertEqual(
            rank_2d(a1, axis=1, method='ordinal', start=1).tolist(),
            [[2, 1], [1, 2], [2, 1], [2, 1], [1, 2]]
        )

    def test_rank_2d_b(self) -> None:
        a1 = np.array([10, 3, -4, 9, 3, -12, 18, 3, 0, 17]).reshape(5,2)
        self.assertEqual(
            rank_2d(a1, axis=0, method='mean', start=1).tolist(),
            [[4.0, 2.5], [1.0, 4.0], [3.0, 1.0], [5.0, 2.5], [2.0, 5.0]]
        )
        self.assertEqual(
            rank_2d(a1, axis=1, method='mean', start=1).tolist(),
            [[2.0, 1.0], [1.0, 2.0], [2.0, 1.0], [2.0, 1.0], [1.0, 2.0]]
        )

    def test_rank_2d_c(self) -> None:
        a1 = np.array([10, 3, -4, 9, 3, -12, 18, 3, 0, 17]).reshape(5,2)
        self.assertEqual(
            rank_2d(a1, axis=0, method='min', start=1).tolist(),
            [[4, 2], [1, 4], [3, 1], [5, 2], [2, 5]]
        )
        self.assertEqual(
            rank_2d(a1, axis=1, method='min', start=1).tolist(),
            [[2, 1], [1, 2], [2, 1], [2, 1], [1, 2]]
        )

    def test_rank_2d_d(self) -> None:
        a1 = np.array([10, 3, -4, 9, 3, -12, 18, 3, 0, 17]).reshape(5,2)
        self.assertEqual(
            rank_2d(a1, axis=0, method='max', start=1).tolist(),
            [[4, 3], [1, 4], [3, 1], [5, 3], [2, 5]]
        )
        self.assertEqual(
            rank_2d(a1, axis=1, method='max', start=1).tolist(),
            [[2, 1], [1, 2], [2, 1], [2, 1], [1, 2]]
        )


    def test_rank_2d_e(self) -> None:
        a1 = np.array([10, 3, -4, 9, 3, -12, 18, 3, 0, 17]).reshape(5,2)
        self.assertEqual(
            rank_2d(a1, axis=0, method='dense', start=1).tolist(),
            [[4, 2], [1, 3], [3, 1], [5, 2], [2, 4]]
        )
        self.assertEqual(
            rank_2d(a1, axis=1, method='dense', start=1).tolist(),
            [[2, 1], [1, 2], [2, 1], [2, 1], [1, 2]]
        )



    #---------------------------------------------------------------------------
    def test_rank_1d_pair_a(self) -> None:
        self.assertEqual(
                rank_1d(np.array([0, 0]), 'mean').tolist(),
                [0.5, 0.5]
                )

        self.assertEqual(
                rank_1d(np.array([0, 0]), 'min').tolist(),
                [0, 0]
                )

        self.assertEqual(
                rank_1d(np.array([0, 0]), 'max').tolist(),
                [1, 1]
                )

        self.assertEqual(
                rank_1d(np.array([0, 0]), 'dense').tolist(),
                [0, 0]
                )

        self.assertEqual(
                rank_1d(np.array([0, 0]), 'ordinal').tolist(),
                [0, 1]
                )

    def test_rank_1d_pair_b(self) -> None:
        self.assertEqual(
                rank_1d(np.array([0, 0, 1]), 'mean').tolist(),
                [0.5, 0.5, 2]
                )
        self.assertEqual(rank_1d(np.array([0, 0, 1]), 'mean').dtype.kind, 'f')

        self.assertEqual(
                rank_1d(np.array([0, 0, 1]), 'min').tolist(),
                [0, 0, 2]
                )
        self.assertEqual(rank_1d(np.array([0, 0, 1]), 'min').dtype.kind, 'i')

        self.assertEqual(
                rank_1d(np.array([0, 0, 1]), 'max').tolist(),
                [1, 1, 2]
                )
        self.assertEqual(rank_1d(np.array([0, 0, 1]), 'max').dtype.kind, 'i')

        self.assertEqual(
                rank_1d(np.array([0, 0, 1]), 'dense').tolist(),
                [0, 0, 1]
                )
        self.assertEqual(rank_1d(np.array([0, 0, 1]), 'dense').dtype.kind, 'i')

        self.assertEqual(
                rank_1d(np.array([0, 0, 1]), 'ordinal').tolist(),
                [0, 1, 2]
                )
        self.assertEqual(rank_1d(np.array([0, 0, 1]), 'ordinal').dtype.kind, 'i')


if __name__ == '__main__':
    unittest.main()




import typing as tp
import unittest
import operator
import os
import sqlite3
import gc

import numpy as np
from hypothesis import given
from arraykit import isna_element

from static_frame.core.frame import Frame
from static_frame.core.frame import FrameGO
from static_frame.core.interface import UFUNC_AXIS_SKIPNA
from static_frame.core.interface import UFUNC_BINARY_OPERATORS
from static_frame.core.interface import UFUNC_UNARY_OPERATORS
from static_frame.core.series import Series
from static_frame.test.property import strategies as sfst
from static_frame.test.test_case import skip_win
from static_frame.test.test_case import temp_file
from static_frame.test.test_case import TestCase


class TestUnit(TestCase):


    @given(sfst.get_frame_or_frame_go())
    def test_basic_attributes(self, f1: Frame) -> None:

        self.assertEqual(len(f1.dtypes), f1.shape[1])
        # self.assertEqual(f1.shape, f1.shape)
        self.assertEqual(f1.ndim, 2)
        # self.assertEqual(f1.unified, len(f1.mloc) <= 1)

        if f1.shape[0] > 0 and f1.shape[1] > 0:
            self.assertTrue(f1.size > 0)
            self.assertTrue(f1.nbytes > 0)
        else:
            self.assertTrue(f1.size == 0)
            self.assertTrue(f1.nbytes == 0)


    @given(sfst.get_frame_or_frame_go(dtype_group=sfst.DTGroup.NUMERIC))
    def test_unary_operators_numeric(self, f1: Frame) -> None:
        for op in UFUNC_UNARY_OPERATORS:
            if op == '__invert__': # invalid on non Boolean
                continue
            func = getattr(operator, op)
            values = f1.values
            # must coerce all blocks to same type to compare to what NP does
            a = func(f1.astype(values.dtype)).values
            b = func(values)
            self.assertAlmostEqualArray(a, b)


    @given(sfst.get_frame_or_frame_go(dtype_group=sfst.DTGroup.BOOL))
    def test_unary_operators_boolean(self, f1: Frame) -> None:
        for op in UFUNC_UNARY_OPERATORS:
            if op != '__invert__': # valid on Boolean
                continue
            func = getattr(operator, op)
            a = func(f1).values
            b = func(f1.values)
            self.assertAlmostEqualArray(a, b)


    @given(sfst.get_frame_or_frame_go(dtype_group=sfst.DTGroup.NUMERIC))
    def test_binary_operators_numeric(self, f1: Frame) -> None:
        for op in UFUNC_BINARY_OPERATORS:
            if op in {
                    '__matmul__',
                    '__pow__',
                    '__lshift__',
                    '__rshift__',
                    '__and__',
                    '__xor__',
                    '__or__',
                    '__mod__',
                    }:
                continue
            func = getattr(operator, op)
            values = f1.values
            # must coerce all blocks to same type to compare to what NP does
            f2 = f1.astype(values.dtype)
            a = func(f2, f2).values
            b = func(values, values)
            self.assertAlmostEqualArray(a, b)


    @given(sfst.get_frame_or_frame_go(dtype_group=sfst.DTGroup.BOOL))
    def test_binary_operators_boolean(self, f1: Frame) -> None:
        for op in UFUNC_BINARY_OPERATORS:
            if op not in {
                    '__and__',
                    '__xor__',
                    '__or__',
                    }:
                continue
            func = getattr(operator, op)
            a = func(f1, f1).values
            values = f1.values
            b = func(values, values)
            self.assertAlmostEqualArray(a, b)

    @given(sfst.get_frame_or_frame_go(
            dtype_group=sfst.DTGroup.NUMERIC,
            index_dtype_group=sfst.DTGroup.STRING,
            min_rows=3,
            max_rows=3,
            min_columns=3,
            max_columns=3)
            )
    def test_binary_operators_matmul(self,
            f1: Frame,
            ) -> None:

        f2 = f1.relabel(columns=f1.index)
        f3 = f2 @ f1
        self.assertAlmostEqualArray(f3.values, f2.values @ f1.values)

    # from hypothesis import reproduce_failure
    # NOTE: was able to improve many of these, but continued to get compliated type cases, and complications
    @given(sfst.get_frame_or_frame_go(
            dtype_group=sfst.DTGroup.NUMERIC_REAL,
            min_rows=1,
            min_columns=1))
    def test_ufunc_axis(self, f1: Frame) -> None:

        for attr, attrs in UFUNC_AXIS_SKIPNA.items():

            if attr in ('std', 'var'):
                continue

            for axis in (0, 1):
                values = f1.values
                # must coerce all blocks to same type to compare to what NP does
                # f2 = f1.astype(values.dtype)

                a = getattr(f1, attr)(axis=axis).values # call the method
                b = attrs.ufunc_skipna(values, axis=axis)

    #             if a.dtype != b.dtype:
    #                 continue
    #             try:
    #                 self.assertAlmostEqualArray(a, b)
    #             except:
    #                 import ipdb; ipdb.set_trace()
    #                 raise

    @given(sfst.get_frame())
    def test_frame_isin(self, f1: Frame) -> None:
        value = f1.iloc[0, 0]
        if (not isna_element(value) and
                not isinstance(value, np.datetime64) and
                not isinstance(value, np.timedelta64)):
            self.assertTrue(f1.isin((value,)).iloc[0, 0])



    # # TODO: intger tests with pow, mod

    #---------------------------------------------------------------------------

    @given(sfst.get_frame_go(), sfst.get_label())
    def test_frame_go_setitem(self, f1: Frame, label: tp.Hashable) -> None:

        shape = f1.shape
        f1['foo'] = label # type: ignore
        self.assertEqual(shape[1] + 1, f1.shape[1])


    @given(sfst.get_arrays_2d_aligned_rows(min_size=2, max_size=2))
    def test_frame_go_extend(self, arrays: tp.Sequence[np.ndarray]) -> None:
        f1 = FrameGO(arrays[0], columns=self.get_letters(arrays[0].shape[1]))
        shape = f1.shape
        f2 = Frame(arrays[1])
        f1.extend(f2)
        self.assertEqual(f1.shape[1], shape[1] + f2.shape[1])


    @given(sfst.get_arrays_2d_aligned_rows(min_size=3))
    def test_frame_go_extend_items(self, arrays: tp.Sequence[np.ndarray]) -> None:
        frame_array = arrays[0]
        # just take first columm form 2d arrays
        series_arrays = [a[:, 0] for a in arrays[1:]]

        f1 = FrameGO(frame_array)
        shape = f1.shape

        letters = self.get_letters(len(series_arrays))

        def items() -> tp.Iterator[tp.Tuple[tp.Hashable, Series]]:
            for idx, label in enumerate(letters):
                s = Series(series_arrays[idx], index=f1.index)
                yield label, s

        f1.extend_items(items())

        self.assertEqual(f1.shape[1], shape[1] + len(series_arrays))


    #---------------------------------------------------------------------------
    # exporters

    @given(sfst.get_frame_or_frame_go())
    def test_frame_to_pairs(self, f1: Frame) -> None:
        for i in range(0, 1):
            post = f1.to_pairs(i)
            if i == 1:
                self.assertEqual(len(post), f1.shape[1]) # type: ignore
            else:
                self.assertEqual(len(post[0][1]), f1.shape[0]) # type: ignore
            self.assertTrue(isinstance(post, tuple))


    @given(sfst.get_frame_or_frame_go(
            dtype_group=sfst.DTGroup.BASIC,
            index_dtype_group=sfst.DTGroup.BASIC,
            ))
    def test_frame_to_pandas(self, f1: Frame) -> None:
        post = f1.to_pandas()
        self.assertTrue(post.shape == f1.shape)
        if not f1.isna().any().any():
            self.assertTrue((post.values == f1.values).all())


    @given(sfst.get_frame_or_frame_go(
            dtype_group=sfst.DTGroup.BASIC,
            index_dtype_group=sfst.DTGroup.BASIC,
            ))
    def test_frame_to_parquet(self, f1: Frame) -> None:
        import pyarrow
        with temp_file('.parquet') as fp:
            try:
                f1.to_parquet(fp)
                self.assertTrue(os.stat(fp).st_size > 0)
            except pyarrow.lib.ArrowNotImplementedError:
                # could be Byte-swapped arrays not supported
                pass


    @given(sfst.get_frame_or_frame_go(
            dtype_group=sfst.DTGroup.CORE,
            index_dtype_group=sfst.DTGroup.CORE,
            ))
    def test_frame_to_msgpack(self, f1: Frame) -> None:
        msg = f1.to_msgpack()

        f2 = Frame.from_msgpack(msg)
        assert f1.equals(f2, compare_name=True, compare_dtype=True, compare_class=True)

        f2 = Frame.from_msgpack(f1.to_msgpack())
        assert f1.equals(f2, compare_name=True, compare_dtype=True, compare_class=True)


    @given(sfst.get_frame_or_frame_go(
            dtype_group=sfst.DTGroup.BASIC,
            index_dtype_group=sfst.DTGroup.BASIC,
            ))
    def test_frame_to_xarray(self, f1: Frame) -> None:
        xa = f1.to_xarray()
        self.assertTrue(tuple(xa.keys()) == tuple(f1.columns))


    @given(sfst.get_frame(
            dtype_group=sfst.DTGroup.BASIC,
            index_dtype_group=sfst.DTGroup.BASIC,
            ))
    def test_frame_to_frame_go(self, f1: Frame) -> None:
        f2 = f1.to_frame_go()
        f2['__new__'] = 10
        self.assertTrue(len(f2.columns) == len(f1.columns) + 1)

    @skip_win  # type: ignore # get UnicodeEncodeError: 'charmap' codec can't encode character '\u0162' in position 0: character maps to <undefined>
    @given(sfst.get_frame_or_frame_go(
            dtype_group=sfst.DTGroup.BASIC,
            ))
    def test_frame_to_csv(self, f1: Frame) -> None:
        with temp_file('.csv') as fp:
            f1.to_csv(fp)
            self.assertTrue(os.stat(fp).st_size > 0)

            # not yet validating result, as edge cases with unusual unicode and non-unique indices are a problem
            # f2 = Frame.from_csv(fp,
            #         index_depth=f1.index.depth,
            #         columns_depth=f1.columns.depth)


    @skip_win  # type: ignore # UnicodeEncodeError
    @given(sfst.get_frame_or_frame_go(
            dtype_group=sfst.DTGroup.BASIC,
            ))
    def test_frame_to_tsv(self, f1: Frame) -> None:
        with temp_file('.txt') as fp:
            f1.to_tsv(fp)
            self.assertTrue(os.stat(fp).st_size > 0)


    @given(sfst.get_frame_or_frame_go(
            dtype_group=sfst.DTGroup.BASIC,
            ))
    def test_frame_to_xlsx(self, f1: Frame) -> None:
        with temp_file('.xlsx') as fp:
            f1.to_xlsx(fp)
            self.assertTrue(os.stat(fp).st_size > 0)

    @given(sfst.get_frame_or_frame_go(
            dtype_group=sfst.DTGroup.BASIC,
            ))
    def test_frame_to_sqlite(self, f1: Frame) -> None:
        with temp_file('.sqlite') as fp:

            try:
                f1.to_sqlite(fp)
                self.assertTrue(os.stat(fp).st_size > 0)
            except (sqlite3.IntegrityError, sqlite3.OperationalError, OverflowError):
                # some indices, after translation, are not unique
                # SQLite is no case sensitive, and does not support unicide
                # OverflowError: Python int too large to convert to SQLite INTEGER
                pass


    @given(sfst.get_frame_or_frame_go(
            dtype_group=sfst.DTGroup.BASIC,
            columns_dtype_group=sfst.DTGroup.STRING,
            index_dtype_group=sfst.DTGroup.STRING
            ))
    def test_frame_to_hdf5(self, f1: Frame) -> None:
        f1 = f1.rename('f1')
        with temp_file('.hdf5') as fp:

            try:
                f1.to_hdf5(fp)
                self.assertTrue(os.stat(fp).st_size > 0)
            except ValueError:
                # will happen for empty strings and unicde that cannot be handled by HDF5
                pass


    @given(sfst.get_frame_or_frame_go())
    def test_frame_to_html(self, f1: Frame) -> None:
        post = f1.to_html()
        self.assertTrue(len(post) > 0)

    @skip_win  # type: ignore # UnicodeEncodeError
    @given(sfst.get_frame_or_frame_go())
    def test_frame_to_html_datatables(self, f1: Frame) -> None:
        post = f1.to_html_datatables(show=False)
        self.assertTrue(len(post) > 0)

    @given(sfst.get_frame_or_frame_go())
    def test_frame_to_rst(self, f1: Frame) -> None:
        post = f1.to_rst()
        self.assertTrue(len(post) > 0)

    @given(sfst.get_frame_or_frame_go())
    def test_frame_to_markdown(self, f1: Frame) -> None:
        post = f1.to_markdown()
        self.assertTrue(len(post) > 0)

    @given(sfst.get_frame_or_frame_go())
    def test_frame_to_latex(self, f1: Frame) -> None:
        post = f1.to_latex()
        self.assertTrue(len(post) > 0)

    @given(sfst.get_frame_or_frame_go())
    def test_frame_blocks_dont_have_reference_cycles(self, f1: Frame) -> None:
        self.assertEqual([f1], gc.get_referrers(f1._blocks))


if __name__ == '__main__':
    unittest.main()





import unittest

import frame_fixtures as ff
import numpy as np

from static_frame import Frame
from static_frame import FrameGO
from static_frame import Series
from static_frame.test.test_case import TestCase


class TestUnit(TestCase):

    def test_frame_via_fill_value_a(self) -> None:

        f1 = ff.parse('s(3,3)|c(I,str)|i(I,str)|v(int)')
        f2 = ff.parse('s(2,2)|c(I,str)|i(I,str)|v(int)')

        f3 = f1.via_fill_value(0) + f2
        self.assertEqual(f3.to_pairs(),
                (('zUvW', (('zUvW', 30205), ('zZbu', -3648), ('ztsv', 91301))), ('zZbu', (('zUvW', 84967), ('zZbu', -176034), ('ztsv', 185734))), ('ztsv', (('zUvW', 5729), ('zZbu', 324394), ('ztsv', -82314))))
                )


        f4 = f1.via_fill_value(0) + f2.iloc[0]
        self.assertEqual(f4.to_pairs(),
                (('zUvW', (('zZbu', -3648), ('ztsv', 91301), ('zUvW', 30205))), ('zZbu', (('zZbu', -176034), ('ztsv', 4850), ('zUvW', -3050))), ('ztsv', (('zZbu', 324394), ('ztsv', 121040), ('zUvW', 167926))))
                )


    def test_frame_via_fill_value_b(self) -> None:

        f1 = ff.parse('s(3,3)|c(I,str)|i(I,str)|v(int)')
        f2 = ff.parse('s(2,2)|c(I,str)|i(I,int)|v(int)') % 3

        f3 =  f1.via_T.via_fill_value(1) * f2.iloc[0]

        self.assertEqual(f3.to_pairs(),
                (('zZbu', (('zUvW', 84967), ('zZbu', 0), ('ztsv', 185734))), ('ztsv', (('zUvW', 5729), ('zZbu', 0), ('ztsv', -82314))), ('zUvW', (('zUvW', 30205), ('zZbu', 0), ('ztsv', 182602))))
                )

        f4 =  f1.via_fill_value(1).via_T * f2.iloc[0]

        self.assertEqual(f4.to_pairs(),
                (('zZbu', (('zUvW', 84967), ('zZbu', 0), ('ztsv', 185734))), ('ztsv', (('zUvW', 5729), ('zZbu', 0), ('ztsv', -82314))), ('zUvW', (('zUvW', 30205), ('zZbu', 0), ('ztsv', 182602))))
                )


    def test_frame_via_fill_value_c(self) -> None:
        f1 = Frame(np.arange(20).reshape(4, 5), index=tuple('abcd'))
        f2 = f1.via_T.via_fill_value(0) * Series((0, 2), index=tuple('bc'))
        self.assertEqual(f2.to_pairs(),
                ((0, (('a', 0), ('b', 0), ('c', 20), ('d', 0))), (1, (('a', 0), ('b', 0), ('c', 22), ('d', 0))), (2, (('a', 0), ('b', 0), ('c', 24), ('d', 0))), (3, (('a', 0), ('b', 0), ('c', 26), ('d', 0))), (4, (('a', 0), ('b', 0), ('c', 28), ('d', 0)))))

    def test_frame_via_fill_value_d(self) -> None:
        f1 = Frame(np.arange(20).reshape(4, 5), index=tuple('abcd'))
        with self.assertRaises(RuntimeError):
            f2 = f1.via_T.via_fill_value(0) * Series((0, 2), index=tuple('bc')).via_fill_value(1)

    def test_frame_via_fill_value_e(self) -> None:

        f1 = FrameGO(index=range(5))
        f1.via_fill_value(0)['a'] = Series([10,20], index=(2,4))
        f1.via_fill_value(-1)['b'] = Series([10,20], index=(0,1))

        self.assertEqual(f1.to_pairs(),
                (('a', ((0, 0), (1, 0), (2, 10), (3, 0), (4, 20))), ('b', ((0, 10), (1, 20), (2, -1), (3, -1), (4, -1))))
                )

        f2 = Frame(index=range(5))
        with self.assertRaises(TypeError):
            f2.via_fill_value(0)['a'] = range(5) # type: ignore #pylint: disable=E1137




if __name__ == '__main__':
    unittest.main()





import typing as tp
import unittest

import numpy as np

# from hypothesis import strategies as st
from hypothesis import given
from hypothesis import settings as hypo_settings

from static_frame.test.property import strategies as sfst

from static_frame.test.test_case import TestCase

from static_frame import TypeBlocks
from static_frame import Index
from static_frame import IndexHierarchy
from static_frame import Series
from static_frame import Frame


class TestUnit(TestCase):


    @given(sfst.get_labels())
    def test_get_labels(self, values: tp.Iterable[tp.Hashable]) -> None:
        for value in values:
            self.assertTrue(isinstance(hash(value), int))

    @given(sfst.get_dtypes())
    def test_get_dtypes(self, dtypes: tp.Iterable[np.dtype]) -> None:
        for dt in dtypes:
            self.assertTrue(isinstance(dt, np.dtype))

    @given(sfst.get_spacing(10))
    def test_get_spacing_10(self, spacing: tp.Iterable[int]) -> None:
        self.assertEqual(sum(spacing), 10)

    @hypo_settings(max_examples=10)  # type: ignore
    @given(sfst.get_shape_1d2d())
    def test_get_shape_1d2d(self, shape: tp.Tuple[int, ...]) -> None:
        self.assertTrue(isinstance(shape, tuple))
        self.assertTrue(len(shape) in (1, 2))

    @hypo_settings(max_examples=10)  # type: ignore
    @given(sfst.get_array_1d2d())
    def test_get_array_1d2d(self, array: np.ndarray) -> None:
        self.assertTrue(isinstance(array, np.ndarray))
        self.assertTrue(array.ndim in (1, 2))

    @hypo_settings(max_examples=10)  # type: ignore
    @given(sfst.get_arrays_2d_aligned_columns(min_size=2))
    def test_get_arrays_2s_aligned_columns(self, arrays: tp.Iterable[np.ndarray]) -> None:
        array_iter = iter(arrays)
        a1 = next(array_iter)
        match = a1.shape[1]
        for array in array_iter:
            self.assertEqual(array.shape[1], match)

    @given(sfst.get_arrays_2d_aligned_rows(min_size=2))
    def test_get_arrays_2s_aligned_rows(self, arrays: tp.Iterable[np.ndarray]) -> None:
        array_iter = iter(arrays)
        a1 = next(array_iter)
        match = a1.shape[0]
        for array in array_iter:
            self.assertEqual(array.shape[0], match)

    @hypo_settings(max_examples=10)  # type: ignore
    @given(sfst.get_blocks())
    def test_get_blocks(self, blocks: tp.Tuple[np.ndarray]) -> None:
        self.assertTrue(isinstance(blocks, tuple))
        for b in blocks:
            self.assertTrue(isinstance(b, np.ndarray))
            self.assertTrue(b.ndim in (1, 2))

    @hypo_settings(max_examples=10)  # type: ignore
    @given(sfst.get_type_blocks())
    def test_get_type_blocks(self, tb: TypeBlocks) -> None:
        self.assertTrue(isinstance(tb, TypeBlocks))
        rows, cols = tb.shape
        col_count = 0
        for b in tb._blocks:
            if b.ndim == 1:
                self.assertEqual(len(b), rows)
                col_count += 1
            else:
                self.assertEqual(b.ndim, 2)
                self.assertEqual(b.shape[0], rows)
                col_count += b.shape[1]

        self.assertEqual(col_count, cols)

    @hypo_settings(max_examples=10) # type: ignore
    @given(sfst.get_index())
    def test_get_index(self, idx: Index) -> None:
        self.assertTrue(isinstance(idx, Index))
        self.assertEqual(len(idx), len(idx.values))

    @hypo_settings(max_examples=10)  # type: ignore
    @given(sfst.get_index_hierarchy())
    def test_get_index_hierarchy(self, idx: IndexHierarchy) -> None:
        self.assertTrue(isinstance(idx, IndexHierarchy))
        self.assertTrue(idx.depth > 1)
        self.assertEqual(len(idx), len(idx.values))

    @hypo_settings(max_examples=10)  # type: ignore
    @given(sfst.get_series())
    def test_get_series(self, series: Series) -> None:
        self.assertTrue(isinstance(series, Series))
        self.assertEqual(len(series), len(series.values))

    @hypo_settings(max_examples=10)  # type: ignore
    @given(sfst.get_frame())
    def test_get_frame(self, frame: Frame) -> None:
        self.assertTrue(isinstance(frame, Frame))
        self.assertEqual(frame.shape, frame.values.shape)

    @hypo_settings(max_examples=10)  # type: ignore
    @given(sfst.get_frame(index_cls=IndexHierarchy, columns_cls=IndexHierarchy))  # type: ignore
    def test_get_frame_hierarchy(self, frame: Frame) -> None:
        self.assertTrue(isinstance(frame, Frame))
        self.assertTrue(frame.index.depth > 1)
        self.assertTrue(frame.columns.depth > 1)
        self.assertEqual(frame.shape, frame.values.shape)


if __name__ == '__main__':
    unittest.main()



from subprocess import PIPE
from subprocess import Popen
from sys import executable

from pytest import mark
import numpy as np
import pandas as pd

import static_frame as sf


try:
    import IPython # pylint: disable=W0611
except ImportError:
    HAS_IPYTHON = False
else:
    HAS_IPYTHON = True


COMMAND = b'(np.__name__, np.__version__, pd.__name__, pd.__version__, sf.__name__, sf.__version__)'


def _test_main(python: str) -> None:

    result = f"{eval(COMMAND.decode(), {'np': np, 'pd': pd, 'sf': sf})}".encode()

    args = (python, '-m', 'static_frame')

    process = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    stdout = process.communicate(COMMAND)[0]

    assert not process.returncode
    # disabling this check as it fails in a TOX context
    #assert result in stdout


def test_main_python() -> None:
    _test_main(executable)


@mark.skipif(not HAS_IPYTHON, reason='Requires IPython.')  # type: ignore
def test_main_ipython() -> None:
    _test_main('ipython')



import typing as tp
from functools import partial
from collections import defaultdict
from itertools import repeat
from itertools import product
from itertools import chain

import numpy as np
from arraykit import resolve_dtype
from arraykit import resolve_dtype_iter

from static_frame.core.index_base import IndexBase
from static_frame.core.index import Index
from static_frame.core.index_hierarchy import IndexHierarchy
from static_frame.core.util import DepthLevelSpecifier
from static_frame.core.util import IndexConstructor
from static_frame.core.util import UFunc
from static_frame.core.util import ufunc_dtype_to_dtype
from static_frame.core.util import ufunc_unique
from static_frame.core.container_util import index_from_optional_constructor
from static_frame.core.type_blocks import TypeBlocks

if tp.TYPE_CHECKING:
    from static_frame.core.frame import Frame #pylint: disable=W0611 #pragma: no cover
    from static_frame.core.series import Series #pylint: disable=W0611 #pragma: no cover



#-------------------------------------------------------------------------------
# for Frame.pivot
def extrapolate_column_fields(
        columns_fields: tp.Sequence[tp.Hashable],
        group: tp.Tuple[tp.Hashable, ...],
        data_fields: tp.Sequence[tp.Hashable],
        func_fields: tp.Iterable[tp.Hashable],
        ) -> tp.Iterable[tp.Hashable]:
    '''"Determine columns to be reatined from gruop and data fields.
    Used in Frame.pivot.

    Args:
        group: a unique label from the the result of doing a group-by with the `columns_fields`.
    '''
    columns_fields_len = len(columns_fields)
    data_fields_len = len(data_fields)

    sub_columns: tp.Iterable[tp.Hashable]

    if columns_fields_len == 1 and data_fields_len == 1:
        if not func_fields:
            sub_columns = group # already a tuple
        else:
            sub_columns = [group + (label,) for label in func_fields]
    elif columns_fields_len == 1 and data_fields_len > 1:
        # create a sub heading for each data field
        if not func_fields:
            sub_columns = list(product(group, data_fields))
        else:
            sub_columns = list(product(group, data_fields, func_fields))
    elif columns_fields_len > 1 and data_fields_len == 1:
        if not func_fields:
            sub_columns = (group,)
        else:
            sub_columns = [group + (label,) for label in func_fields]
    else: # group is already a tuple of the partial column label; need to extend with each data field
        if not func_fields:
            sub_columns = [group + (field,) for field in data_fields]
        else:
            sub_columns = [group + (field, label) for field in data_fields for label in func_fields]

    return sub_columns

def pivot_records_dtypes(
        dtype_map: 'Series',
        data_fields: tp.Iterable[tp.Hashable],
        func_single: tp.Optional[UFunc],
        func_map: tp.Sequence[tp.Tuple[tp.Hashable, UFunc]]
        ) -> tp.Iterator[np.dtype]:
    '''
    Iterator of ordered dtypes, providing multiple dtypes per field when func_map is provided.
    '''
    for field in data_fields:
        dtype = dtype_map[field]
        if func_single:
            yield ufunc_dtype_to_dtype(func_single, dtype)
        else: # we assume
            for _, func in func_map:
                yield ufunc_dtype_to_dtype(func, dtype)

def pivot_records_items(
        blocks: TypeBlocks,
        group_fields_iloc: tp.Iterable[tp.Hashable],
        group_depth: int,
        data_fields_iloc: tp.Iterable[tp.Hashable],
        func_single: tp.Optional[UFunc],
        func_map: tp.Sequence[tp.Tuple[tp.Hashable, UFunc]]
        ) -> tp.Iterator[tp.Tuple[tp.Hashable, tp.Sequence[tp.Any]]]:
    '''
    Given a Frame and pivot parameters, perform the group by ont he group_fields and within each group,
    '''
    # NOTE: this delivers results by label row for use in a Frame.from_records_items constructor
    # take_group_index = group_depth > 1
    # columns_loc_to_iloc = frame.columns._loc_to_iloc

    group_key = group_fields_iloc if group_depth > 1 else group_fields_iloc[0] #type: ignore
    record_size = len(data_fields_iloc) * (1 if func_single else len(func_map))
    record: tp.List[tp.Any]

    for label, _, part in blocks.group(axis=0, key=group_key):
        # label = group_index if take_group_index else group_index[0]
        record = [None] * record_size # This size can be pre allocated,
        pos = 0

        if func_single:
            for column_key in data_fields_iloc:
                values = part._extract_array_column(column_key)
                record[pos] = func_single(values)
                pos += 1
        else:
            for column_key in data_fields_iloc:
                values = part._extract_array_column(column_key)
                for _, func in func_map:
                    record[pos] = func(values)
                    pos += 1

        yield label, record

def pivot_items(
        blocks: TypeBlocks,
        group_fields_iloc: tp.Iterable[tp.Hashable],
        group_depth: int,
        data_field_iloc: tp.Hashable,
        func_single: UFunc,
        ) -> tp.Iterator[tp.Tuple[tp.Hashable, tp.Any]]:
    '''
    Specialized generator of pairs for when we hae only one data_field and one function.
    '''
    group_key = group_fields_iloc if group_depth > 1 else group_fields_iloc[0] #type: ignore

    for label, _, sub in blocks.group(axis=0, key=group_key):
        # label = group if take_group else group[0]
        # will always be first
        values = sub._extract_array_column(data_field_iloc)
        yield label, func_single(values)


def pivot_core(
        *,
        frame: 'Frame',
        index_fields: tp.List[tp.Hashable],
        columns_fields: tp.List[tp.Hashable],
        data_fields: tp.List[tp.Hashable],
        func_fields: tp.Tuple[tp.Hashable, ...],
        func_single: tp.Optional[UFunc],
        func_map: tp.Sequence[tp.Tuple[tp.Hashable, UFunc]],
        fill_value: object = np.nan,
        index_constructor: IndexConstructor = None,
        ) -> 'Frame':
    '''Core implementation of Frame.pivot(). The Frame has already been reduced to just relevant columns, and all fields groups are normalized as lists of hashables.
    '''
    from static_frame.core.series import Series
    from static_frame.core.frame import Frame

    data_fields_len = len(data_fields)
    index_depth = len(index_fields)

    # all are lists of hashables; get converted to lists of integers
    columns_loc_to_iloc = frame.columns._loc_to_iloc
    index_fields_iloc: tp.Sequence[int] = columns_loc_to_iloc(index_fields) #type: ignore
    data_fields_iloc: tp.Sequence[int] = columns_loc_to_iloc(data_fields) #type: ignore
    columns_fields_iloc: tp.Sequence[int] = columns_loc_to_iloc(columns_fields) #type: ignore

    # For data fields, we add the field name, not the field values, to the columns.
    columns_name = tuple(columns_fields)
    if data_fields_len > 1 or not columns_fields:
        # if no columns_fields, have to add values label
        columns_name = tuple(chain(*columns_fields, ('values',)))
    if len(func_map) > 1:
        columns_name = columns_name + ('func',)

    columns_depth = len(columns_name)
    if columns_depth == 1:
        columns_name = columns_name[0] # type: ignore
        columns_constructor = partial(frame._COLUMNS_CONSTRUCTOR, name=columns_name)
    else:
        columns_constructor = partial(frame._COLUMNS_HIERARCHY_CONSTRUCTOR.from_labels,
                depth_reference=columns_depth,
                name=columns_name)

    dtype_map = frame.dtypes
    dtypes_per_data_fields = tuple(pivot_records_dtypes(
            dtype_map=dtype_map,
            data_fields=data_fields,
            func_single=func_single,
            func_map=func_map,
            ))
    if func_single and data_fields_len == 1:
        dtype_single = ufunc_dtype_to_dtype(func_single, dtype_map[data_fields[0]])

    #---------------------------------------------------------------------------
    # first major branch: if we are only grouping be index fields

    if not columns_fields: # group by is only index_fields
        columns = data_fields if func_single else tuple(product(data_fields, func_fields))

        # NOTE: at this time we do not automatically give back an IndexHierarchy when index_depth is == 1, as the order of the resultant values may not be hierarchable.
        name_index = index_fields[0] if index_depth == 1 else tuple(index_fields)
        if index_constructor:
            index_constructor = partial(index_constructor, name=name_index)
        else:
            index_constructor = partial(Index, name=name_index)

        if len(columns) == 1:
            # assert len(data_fields) == 1
            f = frame.from_series(
                    Series.from_items(
                            pivot_items(blocks=frame._blocks,
                                    group_fields_iloc=index_fields_iloc,
                                    group_depth=index_depth,
                                    data_field_iloc=data_fields_iloc[0],
                                    func_single=func_single,
                                    ),
                            name=columns[0],
                            index_constructor=index_constructor,
                            dtype=dtype_single,
                            ),
                    columns_constructor=columns_constructor)
        else:
            f = frame.from_records_items(
                    pivot_records_items(
                            blocks=frame._blocks,
                            group_fields_iloc=index_fields_iloc,
                            group_depth=index_depth,
                            data_fields_iloc=data_fields_iloc,
                            func_single=func_single,
                            func_map=func_map,
                    ),
                    columns_constructor=columns_constructor,
                    columns=columns,
                    index_constructor=index_constructor,
                    dtypes=dtypes_per_data_fields,
                    )

        # have to rename columns if derived in from_concat
        columns_final = (f.columns.rename(columns_name) if columns_depth == 1
                else columns_constructor(f.columns))
        return f.relabel(columns=columns_final) #type: ignore

    #---------------------------------------------------------------------------
    # second major branch: we are only grouping be index and columns fields

    # avoid doing a multi-column-style selection if not needed
    if len(columns_fields) == 1:
        # columns_group = columns_fields[0]
        retuple_group_label = True
    else:
        # columns_group = columns_fields
        retuple_group_label = False

    columns_loc_to_iloc = frame.columns._loc_to_iloc
    # group by on 1 or more columns fields
    # NOTE: explored doing one group on index and coluns that insert into pre-allocated arrays, but that proved slower than this approach
    group_key = columns_fields_iloc if len(columns_fields_iloc) > 1 else columns_fields_iloc[0]

    index_outer = pivot_outer_index(frame=frame,
                index_fields=index_fields,
                index_depth=index_depth,
                index_constructor=index_constructor,
                )

    # collect subframes based on an index of tuples and columns of tuples (if depth > 1)
    sub_blocks = []
    sub_columns_collected: tp.List[tp.Hashable] = []

    # for group, sub in frame.iter_group_items(columns_group):
    for group, _, sub in frame._blocks.group(axis=0, key=group_key):
        # derive the column fields represented by this group
        sub_columns = extrapolate_column_fields(
                columns_fields,
                group if not retuple_group_label else (group,),
                data_fields,
                func_fields)
        sub_columns_collected.extend(sub_columns)

        # sub is TypeBlocks unique value in columns_group; this may or may not have unique index fields; if not, it needs to be aggregated
        if index_depth == 1:
            sub_index_labels = sub._extract_array_column(index_fields_iloc[0])
            sub_index_labels_unique = ufunc_unique(sub_index_labels)
        else: # match to an index of tuples; the order might not be the same as IH
            # NOTE: might be able to keep arays and concat below
            sub_index_labels = tuple(zip(*(
                    sub._extract_array_column(columns_loc_to_iloc(f))
                    for f in index_fields)))
            sub_index_labels_unique = set(sub_index_labels)

        sub_frame: tp.Union[Frame, Series]

        # if sub_index_labels are not unique we need to aggregate
        if len(sub_index_labels_unique) != len(sub_index_labels):
            # if sub_columns length is 1, that means that we only need to extract one column out of the sub Frame
            if len(sub_columns) == 1:
                assert len(data_fields) == 1
                # NOTE: grouping on index_fields; can pre-process array_to_groups_and_locations
                sub_frame = Series.from_items(
                        pivot_items(blocks=sub,
                                group_fields_iloc=index_fields_iloc,
                                group_depth=index_depth,
                                data_field_iloc=data_fields_iloc[0],
                                func_single=func_single,
                                ),
                        dtype=dtype_single,
                        )
            else:
                sub_frame = Frame.from_records_items(
                        pivot_records_items(
                                blocks=sub,
                                group_fields_iloc=index_fields_iloc,
                                group_depth=index_depth,
                                data_fields_iloc=data_fields_iloc,
                                func_single=func_single,
                                func_map=func_map),
                        dtypes=dtypes_per_data_fields,
                        )
        else:
            # we have unique values per index item, but may not have a complete index
            if func_single:
                # NOTE: should apply function even with func_single
                if len(data_fields) == 1:
                    sub_frame = Frame(
                            sub._extract_array_column(data_fields_iloc[0]),
                            index=sub_index_labels,
                            index_constructor=index_constructor,
                            own_data=True)
                else:
                    sub_frame = Frame(
                            sub._extract(row_key=None,
                                    column_key=data_fields_iloc),
                            index=sub_index_labels,
                            index_constructor=index_constructor,
                            own_data=True)
            else:
                def blocks() -> tp.Iterator[np.ndarray]:
                    for field in data_fields_iloc:
                        for _, func in func_map:
                            yield sub._extract_array_column(field)
                sub_frame = Frame(
                        TypeBlocks.from_blocks(blocks()),
                        index=sub_index_labels,
                        own_data=True,
                        )

        sub_frame = sub_frame.reindex(index_outer,
                own_index=True,
                fill_value=fill_value,
                )
        if sub_frame.ndim == 1:
            sub_blocks.append(sub_frame.values)
        else:
            sub_blocks.extend(sub_frame._blocks._blocks) # type: ignore

    tb = TypeBlocks.from_blocks(sub_blocks)
    return frame.__class__(tb,
            index=index_outer,
            columns=columns_constructor(sub_columns_collected),
            own_data=True,
            own_index=True,
            own_columns=True,
            )


#-------------------------------------------------------------------------------

def pivot_outer_index(
        frame: 'Frame',
        index_fields: tp.Sequence[tp.Hashable],
        index_depth: int,
        index_constructor: IndexConstructor = None,
        ) -> IndexBase:

    index_loc = index_fields if index_depth > 1 else index_fields[0]

    if index_depth == 1:
        index_values = ufunc_unique(
                frame._blocks._extract_array_column(
                        frame._columns._loc_to_iloc(index_loc)),
                axis=0)
        name = index_fields[0]
        index_inner = index_from_optional_constructor(
                index_values,
                default_constructor=partial(Index, name=name),
                explicit_constructor=None if index_constructor is None else partial(index_constructor, name=name),
                )
    else: # > 1
        # NOTE: this might force type an undesirable consolidation
        index_values = ufunc_unique(
                frame._blocks._extract_array(
                        column_key=frame._columns._loc_to_iloc(index_loc)),
                axis=0)
        # NOTE: if index_types need to be provided to an IH here, they must be partialed in the single-argument index_constructor
        name = tuple(index_fields)
        index_inner = index_from_optional_constructor( # type: ignore
                index_values,
                default_constructor=partial(
                        IndexHierarchy.from_labels,
                        name=name,
                        ),
                explicit_constructor=None if index_constructor is None else partial(index_constructor, name=name),
                ).flat()
    return index_inner


#-------------------------------------------------------------------------------

class PivotIndexMap(tp.NamedTuple):
    targets_unique: tp.Iterable[tp.Hashable]
    target_depth: int
    target_select: np.ndarray
    group_to_target_map: tp.Dict[tp.Optional[tp.Hashable], tp.Dict[tp.Any, int]]
    group_depth: int
    group_select: np.ndarray
    group_to_dtype: tp.Dict[tp.Optional[tp.Hashable], np.dtype]

def pivot_index_map(*,
        index_src: IndexBase,
        depth_level: DepthLevelSpecifier,
        dtypes_src: tp.Optional[tp.Sequence[np.dtype]],
        ) -> PivotIndexMap:
    '''
    Args:
        dtypes_src: must be of length equal to axis
    '''
    # We are always moving levels from one axis to another; after application, the expanded axis will always be hierarchical, while the contracted axis may or may not be. From the contract axis, we need to divide the depths into two categories: targets (the depths to be moved and added to expand axis) and groups (unique combinations that remain on the contract axis after removing targets).

    # Unique target labels are added to labels on the expand axis; unique group labels become the new contract axis.

    target_select = np.full(index_src.depth, False)
    target_select[depth_level] = True
    group_select = ~target_select

    group_arrays = []
    target_arrays = []
    for i, v in enumerate(target_select):
        if v:
            target_arrays.append(index_src.values_at_depth(i))
        else:
            group_arrays.append(index_src.values_at_depth(i))

    group_depth = len(group_arrays)
    target_depth = len(target_arrays)
    group_to_dtype: tp.Dict[tp.Optional[tp.Hashable], np.dtype] = {}
    targets_unique: tp.Iterable[tp.Hashable]

    if group_depth == 0:
        # targets must be a tuple
        group_to_target_map = {
                None: {v: idx for idx, v in enumerate(zip(*target_arrays))}
                }
        targets_unique = [k for k in group_to_target_map[None]]
        if dtypes_src is not None:
            group_to_dtype[None] = resolve_dtype_iter(dtypes_src)
    else:
        group_to_target_map = defaultdict(dict)
        targets_unique = dict() # Store targets in order observed

        for axis_idx, (group, target, dtype) in enumerate(zip(
                zip(*group_arrays), # get tuples of len 1 to depth
                zip(*target_arrays),
                (dtypes_src if dtypes_src is not None else repeat(None)),
                )):
            if group_depth == 1:
                group = group[0]
            # targets are transfered labels; groups are the new columns
            group_to_target_map[group][target] = axis_idx
            targets_unique[target] = None #type: ignore

            if dtypes_src is not None:
                if group in group_to_dtype:
                    group_to_dtype[group] = resolve_dtype(group_to_dtype[group], dtype)
                else:
                    group_to_dtype[group] = dtype

    return PivotIndexMap( #pylint: disable=E1120
            targets_unique=targets_unique,
            target_depth=target_depth,
            target_select=target_select,
            group_to_target_map=group_to_target_map, #type: ignore
            group_depth=group_depth,
            group_select=group_select,
            group_to_dtype=group_to_dtype
            )


#-------------------------------------------------------------------------------
class PivotDeriveConstructors(tp.NamedTuple):
    contract_dst: tp.Optional[tp.Iterable[tp.Hashable]]
    contract_constructor: IndexConstructor
    expand_constructor: IndexConstructor

def pivot_derive_constructors(*,
        contract_src: IndexBase,
        expand_src: IndexBase,
        group_select: np.ndarray, # Boolean
        group_depth: int,
        target_select: np.ndarray,
        # target_depth: int,
        group_to_target_map: tp.Dict[tp.Hashable, tp.Tuple[tp.Hashable]],
        expand_is_columns: bool,
        frame_cls: tp.Type['Frame'],
        ) -> PivotDeriveConstructors:
    '''
    pivot_stack: columns is contract, index is expand
    pivot_unstack: index is contract, columns is expand
    '''
    # NOTE: group_select, target_select operate on the contract axis
    if expand_is_columns:
        contract_cls = Index
        contract_cls_hierarchy = IndexHierarchy
        expand_cls_hierarchy = frame_cls._COLUMNS_HIERARCHY_CONSTRUCTOR
    else: # contract is columns
        contract_cls = frame_cls._COLUMNS_CONSTRUCTOR
        contract_cls_hierarchy = frame_cls._COLUMNS_HIERARCHY_CONSTRUCTOR
        expand_cls_hierarchy = IndexHierarchy

    # NOTE: not propagating name attr, as not obvious how it should when depths are exiting and entering

    # contract axis may or may not be IndexHierarchy after extracting depths
    if contract_src.depth == 1: # will removed that one level, thus need IndexAuto
        contract_dst = None
        contract_constructor = contract_cls
    else:
        contract_src_types = contract_src.index_types.values
        contract_dst_types = contract_src_types[group_select]
        if group_depth == 0:
            contract_dst = None
            contract_constructor = contract_cls
        elif group_depth == 1:
            contract_dst = list(group_to_target_map.keys())
            contract_constructor = contract_dst_types[0]
        else:
            contract_dst = list(group_to_target_map.keys())
            contract_constructor = partial( #type: ignore
                    contract_cls_hierarchy.from_labels,
                    index_constructors=contract_dst_types,
                    )

    # expand axis will always be IndexHierarchy after adding depth
    if expand_src.depth == 1:
        expand_types = [expand_src.__class__]
    else:
        expand_types = list(expand_src._levels.index_types()) #type: ignore

    if contract_src.depth == 1:
        expand_types.append(contract_src.__class__)
    else:
        expand_types.extend(contract_src_types[target_select])

    expand_constructor = partial(
            expand_cls_hierarchy.from_labels,
            index_constructors=expand_types,
            # name=expand_src.name,
            )

    # NOTE: expand_dst labels will come from the values generator
    return PivotDeriveConstructors( #pylint: disable=E1120
            contract_dst=contract_dst,
            contract_constructor=contract_constructor,
            expand_constructor=expand_constructor,
            )






# import typing as tp
import unittest

# import numpy as np

from hypothesis import given
from static_frame.test.test_case import TestCase

from static_frame.test.property.strategies import get_index_hierarchy_any


from static_frame import IndexHierarchy
# from static_frame import IndexHierarchyGO


class TestUnit(TestCase):

    #---------------------------------------------------------------------------
    @given(get_index_hierarchy_any())
    def test_index_display(self, ih: IndexHierarchy) -> None:

        d1 = ih.display()
        self.assertTrue(len(d1) > 0)

        d2 = ih.display_tall()
        self.assertTrue(len(d2) > 0)

        d3 = ih.display_wide()
        self.assertTrue(len(d3) > 0)

    @given(get_index_hierarchy_any())
    def test_index_to_frame(self, ih: IndexHierarchy) -> None:
        f1 = ih.to_frame()
        self.assertEqual(f1.shape, ih.shape)


if __name__ == '__main__':
    unittest.main()




import typing as tp
import datetime
import unittest
import fractions
import numpy as np

from hypothesis import strategies as st
from hypothesis import given

from arraykit import mloc
from arraykit import shape_filter
from arraykit import resolve_dtype
from arraykit import resolve_dtype_iter
from arraykit import isna_element


# from static_frame.test.property.strategies import get_arrays_2d_aligned
# from static_frame.test.property.strategies import get_blocks
# from static_frame.test.property.strategies import get_label
from static_frame.core import util
from static_frame.core.interface import UFUNC_AXIS_SKIPNA
from static_frame.core.util import DTYPE_INEXACT_KINDS
from static_frame.test.property.strategies import DTGroup
from static_frame.test.property.strategies import get_array_1d
from static_frame.test.property.strategies import get_array_1d2d
from static_frame.test.property.strategies import get_array_2d
from static_frame.test.property.strategies import get_arrays_2d_aligned_columns
from static_frame.test.property.strategies import get_arrays_2d_aligned_rows
from static_frame.test.property.strategies import get_dtype
from static_frame.test.property.strategies import get_dtype_pairs
from static_frame.test.property.strategies import get_dtypes
from static_frame.test.property.strategies import get_labels
from static_frame.test.property.strategies import get_shape_1d2d
from static_frame.test.property.strategies import get_value
from static_frame.test.test_case import TestCase

class TestUnit(TestCase):


    @given(get_array_1d2d())
    def test_mloc(self, array: np.ndarray) -> None:

        x = mloc(array)
        self.assertTrue(isinstance(x, int))

    @given(get_array_1d2d())
    def test_shape_filter(self, shape: np.ndarray) -> None:
        self.assertTrue(len(shape_filter(shape)), 2)

    @given(get_dtype_pairs())
    def test_resolve_dtype(self, dtype_pair: tp.Tuple[np.dtype, np.dtype]) -> None:

        x = resolve_dtype(*dtype_pair)
        self.assertTrue(isinstance(x, np.dtype))

    @given(get_dtypes(min_size=1))
    def test_resolve_dtype_iter(self, dtypes: tp.Iterable[np.dtype]) -> None:

        x = resolve_dtype_iter(dtypes)
        self.assertTrue(isinstance(x, np.dtype))

    @given(get_labels(min_size=1))
    def test_resolve_type_iter(self, objects: tp.Iterable[object]) -> None:

        known_types = set((
                None,
                type(None),
                bool,
                str,
                object,
                int,
                float,
                complex,
                datetime.date,
                datetime.datetime,
                fractions.Fraction
                ))
        resolved, has_tuple, values_post = util.prepare_iter_for_array(objects)
        self.assertTrue(resolved in known_types)



    @given(get_arrays_2d_aligned_columns())
    def test_concat_resolved_axis_0(self, arrays: tp.List[np.ndarray]) -> None:
        array = util.concat_resolved(arrays, axis=0)
        self.assertEqual(array.ndim, 2)
        self.assertEqual(array.dtype, resolve_dtype_iter((x.dtype for x in arrays)))

    @given(get_arrays_2d_aligned_rows())
    def test_concat_resolved_axis_1(self, arrays: tp.List[np.ndarray]) -> None:
        array = util.concat_resolved(arrays, axis=1)
        self.assertEqual(array.ndim, 2)
        self.assertEqual(array.dtype, resolve_dtype_iter((x.dtype for x in arrays)))

    @given(get_dtype(), get_shape_1d2d(), get_value())
    def test_full_or_fill(self,
            dtype: np.dtype,
            shape: tp.Union[tp.Tuple[int], tp.Tuple[int, int]],
            value: object) -> None:
        array = util.full_for_fill(dtype, shape, fill_value=value)
        self.assertTrue(array.shape == shape)
        if isinstance(value, (float, complex)) and np.isnan(value):
            pass
        else:
            self.assertTrue(value in array)

    @given(get_dtype())
    def test_dtype_to_na(self, dtype: util.DtypeSpecifier) -> None:
        post = util.dtype_to_fill_value(dtype)
        self.assertTrue(post in {0, False, None, '', np.nan, util.NAT})


    @given(get_array_1d2d(dtype_group=DTGroup.NUMERIC))
    def test_ufunc_axis_skipna(self, array: np.ndarray) -> None:

        has_na = util.isna_array(array).any()

        for nt in UFUNC_AXIS_SKIPNA.values():
            ufunc = nt.ufunc
            ufunc_skipna = nt.ufunc_skipna
            # dtypes = nt.dtypes
            # composable = nt.composable
            # doc = nt.doc_header
            # size_one_unity = nt.size_one_unity

            with np.errstate(over='ignore', under='ignore', divide='ignore'):

                post = util.array_ufunc_axis_skipna(array=array,
                        skipna=True,
                        axis=0,
                        ufunc=ufunc,
                        ufunc_skipna=ufunc_skipna
                        )
                if array.ndim == 2:
                    self.assertTrue(post.ndim == 1)

    @given(get_array_1d2d())
    def test_ufunc_unique(self, array: np.ndarray) -> None:
        post = util.ufunc_unique(array, axis=0)
        self.assertTrue(len(post) <= array.shape[0])

    @given(get_array_1d(min_size=1), st.integers())
    def test_roll_1d(self, array: np.ndarray, shift: int) -> None:
        post = util.roll_1d(array, shift)
        self.assertEqual(len(post), len(array))
        self.assertEqualWithNaN(array[-(shift % len(array))], post[0])

    @given(get_array_2d(min_rows=1, min_columns=1), st.integers())
    def test_roll_2d(self, array: np.ndarray, shift: int) -> None:
        for axis in (0, 1):
            post = util.roll_2d(array, shift=shift, axis=axis)
            self.assertEqual(post.shape, array.shape)

            start = -(shift % array.shape[axis])

            if axis == 0:
                a = array[start]
                b = post[0]
            else:
                a = array[:, start]
                b = post[:, 0]

            self.assertAlmostEqualValues(a, b)



    @given(get_array_1d(dtype_group=DTGroup.OBJECT))
    def test_iterable_to_array_a(self, array: np.ndarray) -> None:
        values = array.tolist()
        post, _ = util.iterable_to_array_1d(values)
        self.assertAlmostEqualValues(post, values)

        # explicitly giving object dtype
        post, _ = util.iterable_to_array_1d(values, dtype=util.DTYPE_OBJECT)
        self.assertAlmostEqualValues(post, values)


    @given(get_labels())
    def test_iterable_to_array_b(self, labels: tp.Iterable[tp.Any]) -> None:
        post, _ = util.iterable_to_array_1d(labels)
        self.assertAlmostEqualValues(post, labels)
        self.assertTrue(isinstance(post, np.ndarray))


    @given(get_labels())
    def test_iterable_to_array_nd(self, labels: tp.Iterable[tp.Any]) -> None:
        post = util.iterable_to_array_nd(labels)
        self.assertAlmostEqualValues(post, labels)
        self.assertTrue(isinstance(post, np.ndarray))

        if len(labels): #type: ignore
            sample = post[0]
            post = util.iterable_to_array_nd(sample)
            self.assertTrue(isinstance(post, np.ndarray))

    @given(st.slices(10))  #pylint: disable=E1120
    def test_slice_to_ascending_slice(self, key: slice) -> None:

        post_key = util.slice_to_ascending_slice(key, size=10)
        self.assertEqual(
            set(range(*key.indices(10))),
            set(range(*post_key.indices(10)))
            )

# to_datetime64
# to_timedelta64
# key_to_datetime_key

    @given(get_array_1d2d())
    def test_array_to_groups_and_locations(self, array: np.ndarray) -> None:

        groups, locations = util.array_to_groups_and_locations(array, 0)

        if len(array) > 0:
            self.assertTrue(len(groups) >= 1)

        # always 1dm locations
        self.assertTrue(locations.ndim == 1)
        self.assertTrue(len(np.unique(locations)) == len(groups))


    @given(get_array_1d2d())
    def test_isna_array(self, array: np.ndarray) -> None:

        post = util.isna_array(array)
        self.assertTrue(post.dtype == bool)

        values = np.ravel(array)
        count_na = sum(isna_element(x) for x in values)

        self.assertTrue(np.ravel(post).sum() == count_na)


    @given(get_array_1d(dtype_group=DTGroup.BOOL))
    def test_binary_transition(self, array: np.ndarray) -> None:
        post = util.binary_transition(array)

        # could be 32 via result of np.nonzero
        self.assertTrue(post.dtype in (np.int32, np.int64))

        # if no True in original array, result will be empty
        if array.sum() == 0:
            self.assertTrue(len(post) == 0)
        # if all True, result is empty
        elif array.sum() == len(array):
            self.assertTrue(len(post) == 0)
        else:
            # the post selection shold always be indices that are false
            self.assertTrue(array[post].sum() == 0)


    @given(get_array_1d2d())
    def test_array_to_duplicated(self, array: np.ndarray) -> None:
        if array.ndim == 2:
            for axis in (0, 1):
                post = util.array_to_duplicated(array, axis=axis)
                if axis == 0:
                    unique_count = len(set(tuple(x) for x in array))
                else:
                    unique_count = len(set(
                        tuple(array[:, i]) for i in range(array.shape[1]))
                        )
                if unique_count < array.shape[axis]:
                    self.assertTrue(post.sum() > 0)
        else:
            post = util.array_to_duplicated(array)
            # if not all value are unique, we must have some duplicated
            if len(set(array)) < len(array):
                self.assertTrue(post.sum() > 0)

        self.assertTrue(post.dtype == bool)

    @given(get_array_1d2d())
    def test_array_shift(self, array: np.ndarray) -> None:

        for shift in (-1, 1):
            for wrap in (True, False):

                tests = []
                post1 = util.array_shift(
                        array=array,
                        shift=shift,
                        axis=0,
                        wrap=wrap)
                tests.append(post1)

                if array.ndim == 2:
                    post2 = util.array_shift(
                        array=array,
                        shift=shift,
                        axis=1,
                        wrap=wrap)
                    tests.append(post2)

                for post in tests:
                    self.assertTrue(array.shape == post.shape)

                    # type is only always maintained if we are wrapping
                    if wrap:
                        self.assertTrue(array.dtype == post.dtype)


    @given(st.lists(get_array_1d(), min_size=2, max_size=2))
    def test_union1d(self, arrays: tp.Sequence[np.ndarray]) -> None:
        post = util.union1d(
                arrays[0],
                arrays[1],
                assume_unique=False)
        self.assertTrue(post.ndim == 1)

        # the unqiueness of NaNs has changed in newer NP versions, so only compare if non-nans are found
        if post.dtype.kind in ('c', 'f') and not np.isnan(post).any():
            self.assertTrue(len(post) == len(set(arrays[0]) | set(arrays[1])))
        # complex results are tricky to compare after forming sets
        if (post.dtype.kind not in ('O', 'M', 'm', 'c', 'f')
                and not np.isnan(post).any()):
            self.assertSetEqual(set(post), (set(arrays[0]) | set(arrays[1])))


    @given(st.lists(get_array_1d(), min_size=2, max_size=2))
    def test_intersect1d(self, arrays: tp.Sequence[np.ndarray]) -> None:
        post = util.intersect1d(
                arrays[0],
                arrays[1],
                assume_unique=False)
        self.assertTrue(post.ndim == 1)
        # nan values in complex numbers make direct comparison tricky
        self.assertTrue(len(post) == len(set(arrays[0]) & set(arrays[1])))

        if (post.dtype.kind not in ('O', 'M', 'm', 'c', 'f')
                and not np.isnan(post).any()):
            self.assertSetEqual(set(post), (set(arrays[0]) & set(arrays[1])))


    @given(st.lists(get_array_1d(), min_size=2, max_size=2))
    def test_setdiff1d(self, arrays: tp.Sequence[np.ndarray]) -> None:
        post = util.setdiff1d(
                arrays[0],
                arrays[1],
                assume_unique=False)
        self.assertTrue(post.ndim == 1)

        if post.dtype.kind in ('f', 'c', 'i', 'u'):
            # Compare directly to numpy behavior for number values.
            self.assertTrue(len(post) == len(np.setdiff1d(arrays[0], arrays[1], assume_unique=False)))
        else:
            # nan values in complex numbers make direct comparison tricky
            self.assertTrue(len(post) == len(set(arrays[0]).difference(set(arrays[1]))))

        if (post.dtype.kind not in ('O', 'M', 'm', 'c', 'f')
                and not np.isnan(post).any()):
            self.assertSetEqual(set(post), (set(arrays[0]).difference(set(arrays[1]))))


    #---------------------------------------------------------------------------
    @given(get_arrays_2d_aligned_columns(min_size=2, max_size=2))
    def test_union2d(self, arrays: tp.Sequence[np.ndarray]) -> None:
        post = util.union2d(arrays[0], arrays[1], assume_unique=False)
        self.assertTrue(post.ndim == 2)

        if post.dtype.kind in ('f', 'c') and np.isnan(post).any():
            return

        self.assertTrue(len(post) == len(
                set(util.array2d_to_tuples(arrays[0]))
                | set(util.array2d_to_tuples(arrays[1])))
                )

    @given(get_arrays_2d_aligned_columns(min_size=2, max_size=2))
    def test_intersect2d(self, arrays: tp.Sequence[np.ndarray]) -> None:
        post = util.intersect2d(arrays[0], arrays[1], assume_unique=False)
        self.assertTrue(post.ndim == 2)
        self.assertTrue(len(post) == len(
                set(util.array2d_to_tuples(arrays[0]))
                & set(util.array2d_to_tuples(arrays[1])))
                )

    @given(get_arrays_2d_aligned_columns(min_size=2, max_size=2))
    def test_setdiff2d(self, arrays: tp.Sequence[np.ndarray]) -> None:

        for array in arrays:
            if array.dtype.kind in ('f', 'c') and np.isnan(array).any():
                return

        post = util.setdiff2d(arrays[0], arrays[1], assume_unique=False)
        self.assertTrue(post.ndim == 2)
        self.assertTrue(len(post) == len(
                set(util.array2d_to_tuples(arrays[0])).difference(
                set(util.array2d_to_tuples(arrays[1]))))
                )

    @given(get_arrays_2d_aligned_columns())
    def test_array_set_ufunc_many(self, arrays: tp.Sequence[np.ndarray]) -> None:

        for union in (True, False):
            post = util.ufunc_set_iter(arrays, union=union)
            self.assertTrue(post.ndim == 2)

    #---------------------------------------------------------------------------
    @given(get_array_1d2d(min_rows=1, min_columns=1))
    def test_isin(self, array: np.ndarray) -> None:

        container_factory = (list, set, np.array)
        result = None

        if array.ndim == 1:
            sample = array[0]
            if np.array(sample).dtype.kind in DTYPE_INEXACT_KINDS and np.isnan(sample):
                pass
            else:
                for factory in container_factory:
                    result = util.isin(array, factory((sample,)))
                    self.assertTrue(result[0])
        elif array.ndim == 2:
            sample = array[0, 0]
            if np.array(sample).dtype.kind in DTYPE_INEXACT_KINDS and np.isnan(sample):
                pass
            else:
                for factory in container_factory:
                    result = util.isin(array, factory((sample,)))
                    self.assertTrue(result[0, 0])

        if result is not None:
            self.assertTrue(array.shape == result.shape)
            self.assertTrue(result.dtype == bool)




if __name__ == '__main__':
    unittest.main()




import zipfile
import os
import timeit
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import ProcessPoolExecutor
from multiprocessing import get_context as get_mp_context
import shutil
import typing as tp
import gc

import numpy as np
import frame_fixtures as ff
import static_frame as sf
from static_frame.core.display_color import HexColor


COUNT_ARRAY = 100
CHUNK_SIZE = 20

def work(array: np.ndarray):
    # post = []
    # for x in array:
    #     if x > 0:
    #         post.append(x)
    v = array ** 2
    return (v / v.sum()) ** 0.5

class MMapTest:
    NUMBER = 1

    def __init__(self, fixture):
        self.fp_npz = '/tmp/memmap.npz'
        self.fp_dir = '/tmp/memmap'

        self.arrays = {}
        for i in range(COUNT_ARRAY):
            self.arrays[str(i)] = np.arange(1_000_000)

        np.savez(self.fp_npz, **self.arrays)

        # cannot read from zip in mmmap mode, so must extract all
        # https://stackoverflow.com/questions/29080556/how-does-numpy-handle-mmaps-over-npz-files

        with zipfile.ZipFile(self.fp_npz) as zf:
            zf.extractall(self.fp_dir)

        print(self, 'init complete')

    def clear(self) -> None:
        if os.path.exists(self.fp_npz):
            os.unlink(self.fp_npz)
        if os.path.exists(self.fp_dir):
            shutil.rmtree(self.fp_dir)

        print(self, 'del complete')

class MemorySum(MMapTest):

    def __call__(self):
        for a in self.arrays.values():
            work(a)


class MemoryThreadSum(MMapTest):
    def __call__(self):
        with ThreadPoolExecutor() as executor:
            post = tuple(executor.map(work, self.arrays.values(), chunksize=CHUNK_SIZE))
            assert len(post) == COUNT_ARRAY

class MemoryForkSum(MMapTest):

    def __call__(self):
        with ProcessPoolExecutor(mp_context=get_mp_context('fork'), ) as executor:
            post = tuple(executor.map(work, self.arrays.values(), chunksize=CHUNK_SIZE))
            assert len(post) == COUNT_ARRAY

class MemorySpawnSum(MMapTest):

    def __call__(self):
        with ProcessPoolExecutor(mp_context=get_mp_context('spawn')) as executor:
            post = tuple(executor.map(work, self.arrays.values(), chunksize=CHUNK_SIZE))
            assert len(post) == COUNT_ARRAY



class MMapThreadSum(MMapTest):

    @staticmethod
    def func(fp: str):
        a = np.load(fp, mmap_mode='r')
        return work(a)

    def __call__(self):
        fps = (os.path.join(self.fp_dir, f'{fn}.npy') for fn in self.arrays.keys())
        with ThreadPoolExecutor() as executor:
            post = tuple(executor.map(self.func, fps, chunksize=CHUNK_SIZE))
            assert len(post) == COUNT_ARRAY

class MMapForkSum(MMapTest):

    @staticmethod
    def func(fp: str):
        a = np.load(fp, mmap_mode='r')
        return work(a)

    def __call__(self):
        fps = (os.path.join(self.fp_dir, f'{fn}.npy') for fn in self.arrays.keys())
        with ProcessPoolExecutor(mp_context=get_mp_context('fork')) as executor:
            post = tuple(executor.map(self.func, fps, chunksize=CHUNK_SIZE))
            assert len(post) == COUNT_ARRAY

class MMapSpawnSum(MMapTest):

    @staticmethod
    def func(fp: str):
        a = np.load(fp, mmap_mode='r')
        return work(a)

    def __call__(self):
        fps = (os.path.join(self.fp_dir, f'{fn}.npy') for fn in self.arrays.keys())
        with ProcessPoolExecutor(mp_context=get_mp_context('spawn')) as executor:
            post = tuple(executor.map(self.func, fps, chunksize=CHUNK_SIZE))
            assert len(post) == COUNT_ARRAY





#-------------------------------------------------------------------------------
def get_format():

    name_root_last = None
    name_root_count = 0

    def format(key: tp.Tuple[tp.Any, str], v: object) -> str:
        nonlocal name_root_last
        nonlocal name_root_count

        if isinstance(v, float):
            if np.isnan(v):
                return ''
            return str(round(v, 4))
        if isinstance(v, (bool, np.bool_)):
            if v:
                return HexColor.format_terminal('green', str(v))
            return HexColor.format_terminal('orange', str(v))

        return str(v)

    return format


def run_test():
    records = []
    for label, fixture in (
            ('None', None),
            ):
        for cls in (
                MemorySum,
                MemoryThreadSum,
                MemoryForkSum,
                MemorySpawnSum,
                MMapThreadSum,
                MMapForkSum,
                MMapSpawnSum,
                ):
            runner = cls(fixture)
            record = [cls.__name__, cls.NUMBER, label]
            result = timeit.timeit(
                    f'runner()',
                    globals=locals(),
                    number=cls.NUMBER)
            record.append(result)
            records.append(record)
            runner.clear()

    f = sf.FrameGO.from_records(records,
            columns=('name', 'number', 'fixture', 'time')
            )

    display = f.iter_element_items().apply(get_format())

    config = sf.DisplayConfig(
            cell_max_width_leftmost=np.inf,
            cell_max_width=np.inf,
            type_show=False,
            display_rows=200,
            include_index=False,
            )
    print(display.display(config))
    # import ipdb; ipdb.set_trace()



if __name__ == '__main__':
    run_test()







'''

Immutable Data & Immutable Data Frames


Benefits of immutable data

  Safety, or Protection against unintended side effects
    Functions mistakenly mutating their arguments
    Interfaces that hand out references to internally stored data

  Performance: Avoiding copying data and over-allocation
    Functions that make defensive copies of data
    All growable interfaces over-allocate for growth

'''

'''
Native immutable data structures

        String, tuple, frozen set
        namedtuple, frozen dataclass

    Proposals for a frozen dict:

        PEP 416 -- Add a frozendict builtin type | Python.org
        PEP 603 -- Adding a frozenmap type to collections | Python.org

'''
from dataclasses import dataclass

import typing as tp
import numpy as np
import static_frame as sf
import pandas as pd
# 'data'[2] = 20
frozenset

class Data1(tp.NamedTuple):
    square: float
    invert: float

@dataclass(frozen=True)
class Data2:
    square: float
    invert: float


'''

Making NumPy arrays immutable

    NumPy arrays are not growable but are mutable

'''



'''
Immutable DataFrames

    Cannot be done with Pandas

    Initializing a DataFrame with 2D array is perfect example of the dangers

    StaticFrame was motivated by this need
'''

a1 = np.arange(5)
s = pd.Series(a1)
a1[2:] = 2000

a1 = np.arange(20).reshape(4,5)
df = pd.DataFrame(a1, index=tuple('abcd'), columns=tuple('ABCDE'))
a1[:, 3] = 0


'''
Advantages of Immutable Data in StaticFrame

  Safety
    Interfaces can store hand hand out references without fear of mutation
    Initialization from arrays is always safe

  Performance
    Initialization from immutable arrays is no-copy
    Extending and concatenating on aligned indicies is no-copy
'''

def processor(
        functions: tp.Iterable[tp.Callable[[tp.Mapping[str, pd.Series]], pd.Series]],
        init: pd.Series
        ) -> tp.Dict[str, pd.Series]:
    results = {'init': init}

    for func in functions:
        results[func.__name__] = func(results.copy())

    return results


class DataInterface:

    def __init__(self, size: int):
        self._square = pd.Series(np.arange(size) ** 2)
        self._invert = -self._square

    @property
    def square(self) -> pd.Series:
        return self._square.copy()

    @property
    def invert(self) -> pd.Series:
        return self._invert.copy()




'''
Assignment like Moves

    If you have immutable container, how to make changes: create a new container
    With a Series, we allocate a new array
    With a Frame, we only mutate what needs to be mutated
'''

# s = sf.Series(range(6), index=tuple('abcdef'))
# s.assign['c'](300)
# s.assign['c':](300)
# s.assign[['a', 'd', 'f']]((-1, -2, -3))
# s.assign[['a', 'd', 'f']](sf.Series.from_dict(dict(f=-3, d=-2, a=-1)))
# s.assign[s.index.isin(('c', 'f'))](100)


# # Examples of defensive copies

# a1 = np.arange(20).reshape(4,5)
# f = sf.Frame(a1, index=tuple('abcd'), columns=tuple('ABCDE'))

# f.assign['C'](False)
# f.assign.loc['c', 'B'](-20)
# f.assign.loc['c', 'B':](-20)
# f.assign.loc['c':, 'B':](False)
# f.assign.loc['c':, 'B':].apply(lambda f: f*1000)









import typing as tp
from functools import partial
from copy import deepcopy

from static_frame.core.index_hierarchy import IndexHierarchy
from static_frame.core.bus import Bus
from static_frame.core.index_base import IndexBase
from static_frame.core.index_level import TreeNodeT
from static_frame.core.exception import AxisInvalid
from static_frame.core.util import AnyCallable
from static_frame.core.util import array_deepcopy

if tp.TYPE_CHECKING:
    from static_frame.core.yarn import Yarn #pylint: disable=W0611 #pragma: no cover


def get_extractor(
        deepcopy_from_bus: bool,
        is_array: bool,
        memo_active: bool,
        ) -> AnyCallable:
    '''
    Args:
        memo_active: enable usage of a common memoization dictionary accross all calls to extract from this extractor.
    '''
    if deepcopy_from_bus:
        memo: tp.Optional[tp.Dict[int, tp.Any]] = None if not memo_active else {}
        if is_array:
            return partial(array_deepcopy, memo=memo)
        return partial(deepcopy, memo=memo)
    return lambda x: x


def bus_to_hierarchy(
        bus: tp.Union[Bus, 'Yarn'],
        axis: int,
        deepcopy_from_bus: bool,
        init_exception_cls: tp.Type[Exception],
        ) -> tp.Tuple[IndexHierarchy, IndexBase]:
    '''
    Given a :obj:`Bus` and an axis, derive a :obj:`IndexHierarchy`; also return and validate the :obj:`Index` of the opposite axis.
    '''
    # NOTE: need to extract just axis labels, not the full Frame; need new Store/Bus loaders just for label data
    extractor = get_extractor(deepcopy_from_bus, is_array=False, memo_active=False)

    def tree_extractor(index: IndexBase) -> tp.Union[IndexBase, TreeNodeT]:
        index = extractor(index)
        if isinstance(index, IndexHierarchy):
            return index.to_tree()
        return index

    tree: TreeNodeT = {}
    opposite: tp.Optional[IndexBase] = None

    for label, f in bus.items():
        if axis == 0:
            tree[label] = tree_extractor(f.index)
            if opposite is None:
                opposite = extractor(f.columns)
            else:
                if not opposite.equals(f.columns):
                    raise init_exception_cls('opposite axis must have equivalent indices')
        elif axis == 1:
            tree[label] = tree_extractor(f.columns)
            if opposite is None:
                opposite = extractor(f.index)
            else:
                if not opposite.equals(f.index):
                    raise init_exception_cls('opposite axis must have equivalent indices')
        else:
            raise AxisInvalid(f'invalid axis {axis}')

    return IndexHierarchy.from_tree(tree), opposite # type: ignore


def buses_to_hierarchy(
        buses: tp.Iterable[Bus],
        labels: tp.Iterable[tp.Hashable],
        deepcopy_from_bus: bool,
        init_exception_cls: tp.Type[Exception],
        ) -> IndexHierarchy:
    '''
    Given an iterable of named :obj:`Bus` derive a :obj:`Series` with an :obj:`IndexHierarchy`.
    '''
    # NOTE: for now, the Returned Series will have bus Names as values; this requires the Yarn to store a dict, not a list

    extractor = get_extractor(deepcopy_from_bus, is_array=False, memo_active=False)

    tree = {}
    for label, bus in zip(labels, buses):
        if not isinstance(bus, Bus):
            raise init_exception_cls(f'Must provide an interable of Bus.')
        if label in tree:
            raise init_exception_cls(f'Bus names must be unique: {label} duplicated')
        tree[label] = extractor(bus._index)

    return IndexHierarchy.from_tree(tree)





import os
import timeit
import tempfile
import typing as tp
import pickle

import numpy as np
import frame_fixtures as ff
import static_frame as sf
from static_frame.core.display_color import HexColor


FF_wide = 's(10,1_000)|v(int,int,bool,float,float)|i(I,str)|c(I,int)'
FF_tall = 's(1_000,10)|v(int,int,bool,float,float)|i(I,str)|c(I,int)'
FF_square = 's(100,100)|v(float)|i(I,str)|c(I,int)'



class FileIOTest:
    NUMBER = 4
    SUFFIX = '.zip'

    def __init__(self, fixtures: tp.Iterable[str]):
        # TODO: store fixtures
        frames = []
        for i, fixture in enumerate(fixtures):
            frames.append(ff.parse(fixture).rename(str(i)))
        self.fixtures = tuple(frames)

        _, self.fp = tempfile.mkstemp(suffix=self.SUFFIX)

    def __del__(self) -> None:
        os.unlink(self.fp)

    def __call__(self):
        raise NotImplementedError()


class FileReadBusParquet(FileIOTest):
    def __init__(self, fixtures: tp.Iterable[str]):
        super().__init__(fixtures)
        b = sf.Bus.from_frames(self.fixtures)
        b.to_zip_parquet(self.fp)
        self.config = sf.StoreConfig(index_depth=1, columns_depth=1)

    def __call__(self):
        b = sf.Bus.from_zip_parquet(self.fp, config=self.config)
        tuple(b.items())

class FileWriteBusParquet(FileIOTest):
    def __init__(self, fixtures: tp.Iterable[str]):
        super().__init__(fixtures)
        self.b = sf.Bus.from_frames(self.fixtures)

    def __call__(self):
        self.b.to_zip_parquet(self.fp)




class FileReadBusNPZ(FileIOTest):
    def __init__(self, fixtures: tp.Iterable[str]):
        super().__init__(fixtures)
        b = sf.Bus.from_frames(self.fixtures)
        b.to_zip_npz(self.fp)

    def __call__(self):
        b = sf.Bus.from_zip_npz(self.fp)
        tuple(b.items())

class FileWriteBusNPZ(FileIOTest):
    def __init__(self, fixtures: tp.Iterable[str]):
        super().__init__(fixtures)
        self.b = sf.Bus.from_frames(self.fixtures)

    def __call__(self):
        self.b.to_zip_npz(self.fp)



def get_format():

    name_root_last = None
    name_root_count = 0

    def format(key: tp.Tuple[tp.Any, str], v: object) -> str:
        nonlocal name_root_last
        nonlocal name_root_count

        if isinstance(v, float):
            if np.isnan(v):
                return ''
            return str(round(v, 4))
        if isinstance(v, (bool, np.bool_)):
            if v:
                return HexColor.format_terminal('green', str(v))
            return HexColor.format_terminal('orange', str(v))

        return str(v)

    return format


def run_test():
    records = []
    for label, fixture in (
            ('wide', FF_wide),
            ('tall', FF_tall),
            ('square', FF_square),
            ):
    # for label, fixture in (('square', FF_square),):
        fixtures = [fixture] * 100
        for cls in (
                FileWriteBusParquet,
                FileWriteBusNPZ,
                # FileWritePickle,
                FileReadBusParquet,
                FileReadBusNPZ,
                # FileReadPickle,
                ):
            runner = cls(fixtures)
            record = [cls.__name__, cls.NUMBER, label]
            result = timeit.timeit(
                    f'runner()',
                    globals=locals(),
                    number=cls.NUMBER)
            record.append(result)
            records.append(record)

    f = sf.FrameGO.from_records(records,
            columns=('name', 'number', 'fixture', 'time')
            )

    display = f.iter_element_items().apply(get_format())

    config = sf.DisplayConfig(
            cell_max_width_leftmost=np.inf,
            cell_max_width=np.inf,
            type_show=False,
            display_rows=200,
            include_index=False,
            )
    print(display.display(config))
    # import ipdb; ipdb.set_trace()

if __name__ == '__main__':
    run_test()


